import os

from vivarium.library.units import units
from vivarium_cell.library.fasta import read_sequence
from vivarium_cell.library.polymerize import generate_template
from vivarium_cell.data.knowledge_base import KnowledgeBase
from vivarium_cell.states.chromosome import Chromosome


ECOLI_GENOME_FILE_NAME = 'Escherichia_coli_str_k_12_substr_mg1655.ASM584v2.dna.chromosome.Chromosome.fa'
ECOLI_GENOME_PATH = os.path.abspath(
    os.path.join(
        __file__,
        "../../flat/{}".format(ECOLI_GENOME_FILE_NAME)
    )
)


class FlagellaChromosome(object):
    ecoli_sequence = None
    knowledge_base = None

    def __init__(self, parameters={}):
        if self.knowledge_base is None:
            self.knowledge_base = KnowledgeBase()
        if self.ecoli_sequence is None:
            self.ecoli_sequence = read_sequence(ECOLI_GENOME_PATH)

        self.factor_thresholds = {
            ('flhDp', 'CRP'): 1e-05 * units.mM,
            ('fliLp1', 'flhDC'): 1e-06 * units.mM,
            ('fliLp1', 'fliA'): 1.3e-05 * units.mM,
            ('fliEp1', 'flhDC'): 4e-06 * units.mM,
            ('fliEp1', 'fliA'): 1.1e-05 * units.mM,
            ('fliFp1', 'flhDC'): 7e-06 * units.mM,
            ('fliFp1', 'fliA'): 1e-05 * units.mM,
            ('flgBp', 'flhDC'): 1e-05 * units.mM,
            ('flgBp', 'fliA'): 8e-06 * units.mM,
            ('flgAp', 'flhDC'): 1.3e-05 * units.mM,
            ('flgAp', 'fliA'): 6e-06 * units.mM,
            ('flhBp', 'flhDC'): 1.5e-05 * units.mM,
            ('flhBp', 'fliA'): 5e-06 * units.mM,
            ('fliAp1', 'flhDC'): 1.7e-05 * units.mM,
            ('fliAp1', 'fliA'): 4e-06 * units.mM,
            ('flgEp', 'flhDC'): 1.9e-05 * units.mM,
            ('flgEp', 'fliA'): 3e-06 * units.mM,
            ('fliDp', 'flhDC'): 1.9e-05 * units.mM,
            ('fliDp', 'fliA'): 3e-06 * units.mM,
            ('flgKp', 'flhDC'): 2.1e-05 * units.mM,
            ('flgKp', 'fliA'): 1e-06 * units.mM,
            ('fliCp', 'fliA'): 5e-06 * units.mM,
            ('tarp', 'fliA'): 7e-06 * units.mM,
            ('motAp', 'fliA'): 9e-06 * units.mM,
            ('flgMp', 'fliA'): 1.1e-06 * units.mM}

        self.factor_thresholds.update(parameters.get('thresholds', {}))

        self.chromosome_config = {
            'sequence': self.ecoli_sequence,
            'genes': {
                'flhDC': ['flhD', 'flhC'],
                'fliL': ['fliL', 'fliM', 'fliN', 'fliO', 'fliP', 'fliQ', 'fliR'],
                'fliE': ['fliE'],
                'fliF': ['fliF', 'fliG', 'fliH', 'fliI', 'fliJ', 'fliK'],
                'flgA': ['flgA', 'flgM', 'flgN'],
                'flgM': ['flgM', 'flgN'],
                'flgE': ['flgE'],
                'flgB': ['flgB', 'flgC', 'flgD', 'flgE', 'flgF', 'flgG', 'flgH', 'flgI', 'flgJ'],
                'flhB': ['flhB', 'flhA', 'flhE'],
                'fliA': ['fliA', 'fliZ'], # ignore 'tcyJ' for now
                'fliD': ['fliD', 'fliS', 'fliT'],
                'flgK': ['flgK', 'flgL'],
                'fliC': ['fliC'],
                'tar': ['tar', 'tap', 'cheR', 'cheB', 'cheY', 'cheZ'],
                'motA': ['motA', 'motB', 'cheA', 'cheW']},
            'promoters': {
                'flhDp': {
                    'id': 'flhDp',
                    'position': 1978197,
                    'direction': -1,
                    'sites': [{'thresholds': {'CRP': 1e-05}}],
                    'terminators': [
                        {
                            'position': 1977266,
                            'strength': 1.0,
                            'products': ['flhDC']}]},
                'fliLp1': {
                    'id': 'fliLp1',
                    'position': 2019618,
                    'direction': 1,
                    'sites': [
                        {'thresholds': {'flhDC': 4e-06}},
                        {'thresholds': {'fliA': 1.3e-05}}],
                    'terminators': [
                        {
                            'position': 2023678,
                            'strength': 1.0,
                            'products': ['fliL']}]},
                'fliEp1': {
                    'id': 'fliEp1',
                    'position': 2013014,
                    'direction': -1,
                    'sites': [
                        {'thresholds': {'flhDC': 5e-06}},
                        {'thresholds': {'fliA': 1.1e-05}}],
                    'terminators': [
                        {
                            'position': 2012700,
                            'strength': 1.0,
                            'products': ['fliE']}]},
                'fliFp1': {
                    'id': 'fliFp1',
                    'position': 2013229,
                    'direction': 1,
                    'sites': [
                        {'thresholds': {'flhDC': 7e-06}},
                        {'thresholds': {'fliA': 1e-05}}],
                    'terminators': [
                        {
                            'position': 2019513,
                            'strength': 1.0,
                            'products': ['fliF']}]},
                'flgBp': {
                    'id': 'flgBp',
                    'position': 1130863,
                    'direction': -1,
                    'sites': [
                        {'thresholds': {'flhDC': 1e-05}},
                        {'thresholds': {'fliA': 8e-06}}],
                    'terminators': [
                        {
                            'position': 1129414,
                            'strength': 1.0,
                            'products': ['flgA']}]},
                'flgAp': {
                    'id': 'flgAp',
                    'position': 1131018,
                    'direction': 1,
                    'sites': [
                        {'thresholds': {'flhDC': 1.3e-05}},
                        {'thresholds': {'fliA': 6e-06}}],
                    'terminators': [
                        {
                            'position': 1138312,
                            'strength': 1.0,
                            'products': ['flgB']}]},
                'flhBp': {
                    'id': 'flhBp',
                    'position': 1966191,
                    'direction': -1,
                    'sites': [
                        {'thresholds': {'flhDC': 1.5e-05}},
                        {'thresholds': {'fliA': 5e-06}}],
                    'terminators': [
                        {
                            'position': 1962580,
                            'strength': 1.0,
                            'products': ['flhB']}]},
                'fliAp1': {
                    'id': 'fliAp1',
                    'position': 2001789,
                    'direction': -1,
                    'sites': [
                        {'thresholds': {'flhDC': 1.7e-05}},
                        {'thresholds': {'fliA': 4e-06}}],
                    'terminators': [
                        {
                            'position': 1999585,
                            'strength': 1.0,
                            'products': ['fliA']}]},
                'flgEp': {
                    'id': 'flgEp',
                    'position': 1132574,
                    'direction': 1,
                    'sites': [
                        {'thresholds': {'flhDC': 1.9e-05}},
                        {'thresholds': {'fliA': 3e-06}}],
                    'terminators': [
                        {
                            'position': 1133782,
                            'strength': 1.0,
                            'products': ['flgE']}]},
                'fliDp': {
                    'id': 'fliDp',
                    'position': 2003872,
                    'direction': 1,
                    'sites': [
                        {'thresholds': {'flhDC': 1.9e-05}},
                        {'thresholds': {'fliA': 3e-06}}],
                    'terminators': [
                        {
                            'position': 2006078,
                            'strength': 1.0,
                            'products': ['fliD']}]},
                'flgKp': {
                    'id': 'flgKp',
                    'position': 1138378,
                    'direction': 1,
                    'sites': [
                        {'thresholds': {'flhDC': 2.1e-05}},
                        {'thresholds': {'fliA': 1e-06}}],
                    'terminators': [
                        {
                            'position': 1140986,
                            'strength': 1.0,
                            'products': ['flgK']}]},
                'fliCp': {
                    'id': 'fliCp',
                    'position': 2002110,
                    'direction': 1,
                    'sites': [
                        {'thresholds': {'fliA': 5e-06}}],
                        # {'thresholds': {'GadE': 0.55, 'H-NS': 0.6}}],
                    'terminators': [
                        {
                            'position': 2003606,
                            'strength': 1.0,
                            'products': ['fliC']}]},
                'tarp': {
                    'id': 'tarp',
                    'position': 1972691,
                    'direction': -1,
                    'sites': [{'thresholds': {'fliA': 7e-06}}],
                        # {'thresholds': {'Fnr': 1e-5}}
                    'terminators': [
                        {
                            'position': 1971030,
                            'strength': 1.0,
                            'products': ['tar']}]},
                'motAp': {
                    'id': 'motAp',
                    'position': 1977139,
                    'direction': -1,
                    'sites': [{'thresholds': {'fliA': 9e-06}}],
                        # {'thresholds': {'CpxR': 1e-5}}],
                    'terminators': [
                        {
                            'position': 1976252,
                            'strength': 1.0,
                            'products': ['motA']}]},
                'flgMp': {
                    'id': 'flgMp',
                    'position': 1130128,
                    'direction': -1,
                    'sites': [{'thresholds': {'fliA': 1.1e-05}}],
                        # {'thresholds': {'CsgD': 0.1}}
                    'terminators': [
                        {
                            'position': 1129414,
                            'strength': 1.0,
                            'products': ['flgM']}]}},
            'domains': {
                0: {
                    'id': 0,
                    'lead': 0,
                    'lag': 0,
                    'children': []}},
            'rnaps': {}}

        # build chromosome and apply thresholds
        self.chromosome = Chromosome(self.chromosome_config)
        self.chromosome.apply_thresholds(self.factor_thresholds)
        self.chromosome_config['promoters'] = {
            key: promoter.to_dict()
            for key, promoter in self.chromosome.promoters.items()}

        self.promoters = [
            'flhDp', 'fliLp1', 'fliEp1', 'fliFp1', 'flgAp', 'flgBp', 'flhBp',
            'fliAp1', 'fliDp', 'flgKp', 'fliCp', 'tarp', 'motAp' 'flgMp']

        self.flhDC_activated = [
            'fliLp1', 'fliEp1', 'fliFp1', 'flgAp', 'flgBp', 'flgEp', 'flhBp',
            'fliAp1', 'fliDp', 'flgKp']

        self.fliA_activated = [
            'fliCp', 'tarp', 'motAp', 'flgMp']

        flhDC_factors = {
            'fliLp1': {
                'flhDC': 1.2, 'fliA': 0.25},
            'fliEp1': {
                'flhDC': 0.45, 'fliA': 0.35},
            'fliFp1': {
                'flhDC': 0.35, 'fliA': 0.30},
            'flgBp': {
                'flhDC': 0.35, 'fliA': 0.45},
            'flgAp': {
                'flhDC': 0.15, 'fliA': 0.3},
            'flgEp': {
                'flhDC': 1.0, 'fliA': 4.0},
            'flhBp': {
                'flhDC': 0.1, 'fliA': 0.35},
            'fliAp1': {
                'flhDC': 1.0, 'fliA': 0.3},
            'fliDp': {
                'flhDC': 1.2, 'fliA': 0.25},
            'flgKp': {
                'flhDC': 1.2, 'fliA': 0.25}
        }

        def binary_sum_gates(promoter_factors):
            affinities = {}
            first, second = list(promoter_factors[
                list(promoter_factors.keys())[0]].keys())

            # this hard coding of simple addition is alarming and probably points
            # towards providing a function of promoter state for affinity rather
            # than a simple lookup of the affinity for each promoter state tuple.
            for promoter, factors in promoter_factors.items():
                affinities[(promoter, first, None)] = factors[first]
                affinities[(promoter, None, second)] = factors[second]
                affinities[(promoter, first, second)] = factors[first] + factors[second]

            return affinities

        # promoter affinities are binding affinity of RNAP onto promoter
        self.promoter_affinities = {
            ('flhDp', 'CRP'): 0.01}
        # self.promoter_affinities[('motAp', 'CpxR')] = 1.0
        flhDC_affinities = binary_sum_gates(flhDC_factors)
        self.promoter_affinities.update(flhDC_affinities)
        # for promoter in self.flhDC_activated:
        #     self.promoter_affinities[(promoter, 'flhDC')] = 1.0
        for promoter in self.fliA_activated:
            self.promoter_affinities[(promoter, 'fliA')] = 1.0
        self.promoter_affinities.update(
            parameters.get('promoter_affinities', {}))
        promoter_affinity_scaling = parameters.get('promoter_affinity_scaling', 1)
        self.promoter_affinities = {
            promoter: affinity * promoter_affinity_scaling
            for promoter, affinity in self.promoter_affinities.items()}

        self.transcripts = [
            (operon, product)
            for operon, products in self.chromosome_config['genes'].items()
            for product in products]

        self.protein_sequences = {
            (operon, product): self.knowledge_base.proteins[
                self.knowledge_base.genes[product]['id']]['seq']
            for operon, product in self.transcripts}

        self.transcript_templates = {
            key: generate_template(
                key,
                len(sequence),
                [key[1]])
            for key, sequence in self.protein_sequences.items()}

        # transcript affinities are the affinities transcripts to bind a ribosome and translate to protein
        # transcript affinities are scaled relative to the requirements to build a single full flagellum.
        self.min_tr_affinity = parameters.get('min_tr_affinity', 1e-1)
        tr_affinity_scaling = {
            'fliL': 2,
            'fliM': 34,
            'fliG': 26,
            'fliH': 12,
            'fliI': 6,
            'fliD': 5,
            'flgE': 120}
        self.transcript_affinities = {}
        for (operon, product) in self.transcripts:
            self.transcript_affinities[(operon, product)] = self.min_tr_affinity * tr_affinity_scaling.get(product,1)
        self.transcript_affinities.update(
            parameters.get('transcript_affinities', {}))


        self.transcription_factors = [
            'flhDC', 'fliA', 'CsgD', 'CRP', 'GadE', 'H-NS', 'CpxR', 'Fnr']

        self.complexation_monomer_ids = [
            'fliG', 'fliM', 'fliN', 'flhA', 'flhB', 'flhD', 'flhC', 'fliO',
            'fliP', 'fliQ', 'fliR', 'fliJ', 'fliI', 'fliH', 'fliL', 'flgH',
            'motA', 'motB', 'flgB', 'flgC', 'flgF', 'flgG', 'flgI', 'fliF',
            'fliE','fliC','flgL','flgK','fliD','flgE']

        self.complexation_complex_ids = [
            'flhDC',
            'flagellar motor switch',
            'flagella',
            'flagellar export apparatus subunit',
            'flagellar export apparatus',
            'flagellar hook',
            'flagellar motor']

        self.complexation_stoichiometry = {
            'flhDC': {
                'flhD': -4.0,
                'flhC': -2.0,
                'flhDC': 1.0
            },
            'flagellar motor switch reaction': {
                'flagellar motor switch': 1.0,
                'fliG': -26.0,
                'fliM': -34.0,
                'fliN': -1.0
            },
            'flagellar export apparatus reaction 1': {
                'flagellar export apparatus subunit': 1.0,
                'flhA': -1.0,
                'flhB': -1.0,
                'fliO': -1.0,
                'fliP': -1.0,
                'fliQ': -1.0,
                'fliR': -1.0,
                'fliJ': -1.0,
                'fliI': -6.0
            },
            'flagellar export apparatus reaction 2': {
                'flagellar export apparatus': 1.0,
                'flagellar export apparatus subunit': -1.0,
                'fliH': -12.0,
            },
            'flagellar motor reaction': {
                'flagellar motor': 1.0,
                'flagellar motor switch': -1.0,
                'fliL': -2.0,
                'flgH': -1.0,
                'motA': -1.0,
                'motB': -1.0,
                'flgB': -1.0,
                'flgC': -1.0,
                'flgF': -1.0,
                'flgG': -1.0,
                'flgI': -1.0,
                'fliF': -1.0,
                'fliE': -1.0,
            },
            'flagellar hook reaction': {
                'flagellar hook': 1,
                'flgE': -120.0,
            },
            'flagellum reaction': {
                'flagella': 1.0,
                'flagellar export apparatus': -1.0,
                'flagellar motor': -1.0,
                'fliC': -1.0,
                'flgL': -1.0,
                'flgK': -1.0,
                'fliD': -5.0,
                'flagellar hook': -1,
            }
        }

        reaction_default = 1e-4
        self.complexation_rates = {
            'flhDC': reaction_default,
            'flagellar motor switch reaction': reaction_default,
            'flagellar export apparatus reaction 1': reaction_default,
            'flagellar export apparatus reaction 2': reaction_default,
            'flagellar motor reaction': reaction_default,
            'flagellar hook reaction': reaction_default,
            'flagellum reaction': reaction_default}




'''
Kinetic rate law generation using the Convenience Kinetics formulation of Michaelis-Menten kinetics

Formulation provided in:
    Liebermeister, Wolfram, and Edda Klipp. "Bringing metabolic networks to life:
    convenience rate law and thermodynamic constraints."
    Theoretical Biology and Medical Modelling 3.1 (2006): 41.

# TODO -- make a vmax options if enzyme kcats not available
'''

from __future__ import absolute_import, division, print_function

import os

import numpy as np

from vivarium.library.dict_utils import tuplify_port_dicts



def get_molecules(reactions):
    '''
    Get a list of all molecules used by reactions

    Args:
           reaction (dict): all reactions that will be used by transport
    Returns:
           self.molecule_ids (list): all molecules used by these reactions
    '''

    molecule_ids = []
    for reaction_id, specs in reactions.items():
        stoichiometry = specs['stoichiometry']
        substrates = stoichiometry.keys()
        enzymes = specs['catalyzed by']
        # Add all relevant molecules_ids
        molecule_ids.extend(substrates)
        molecule_ids.extend(enzymes)
    return list(set(molecule_ids))

# Helper functions
def make_configuration(reactions):
    '''
    Make the rate law configuration, which tells the parameters where to be placed.

    Args:
        reactions (dict): all reactions that will be made into rate laws, in the same format as all_reactions (above).

    Returns:
        rate_law_configuration (dict): includes partition and reaction_cofactor entries for each reaction
    '''

    rate_law_configuration = {}
    # gets all potential interactions between the reactions
    for reaction_id, specs in reactions.items():
        enzymes = specs['catalyzed by']
        # initialize all enzymes
        for enzyme in enzymes:
            if enzyme not in rate_law_configuration:
                rate_law_configuration[enzyme] = {
                    'partition': [],
                    'reaction_cofactors': {},
                }

    # identify parameters for reactions
    for reaction_id, specs in reactions.items():
        stoich = specs.get('stoichiometry')
        enzymes = specs.get('catalyzed by', None)
        reversibility = specs.get('is reversible', False)

        # get sets of cofactors driving this reaction
        forward_cofactors = [mol for mol, coeff in stoich.items() if coeff < 0]
        cofactors = [forward_cofactors]

        if reversibility:
            reverse_cofactors = [mol for mol, coeff in stoich.items() if coeff > 0]
            cofactors.append(reverse_cofactors)

        # get partition, reactions, and parameter indices for each enzyme, and save to rate_law_configuration dictionary
        for enzyme in enzymes:

            # get competition for this enzyme from all other reactions
            competing_reactions = [rxn for rxn, specs2 in reactions.items() if
                    (rxn is not reaction_id) and (enzyme in specs2['catalyzed by'])]

            competitors = []
            for reaction2 in competing_reactions:
                stoich2 = reactions[reaction2]['stoichiometry']
                reactants2 = [mol for mol, coeff in stoich2.items() if coeff < 0]
                competitors.append(reactants2)

            # partition includes both competitors and cofactors.
            partition = competitors + cofactors
            rate_law_configuration[enzyme]['partition'] = partition
            rate_law_configuration[enzyme]['reaction_cofactors'][reaction_id] = cofactors

    return rate_law_configuration

def cofactor_numerator(concentration, km):
    return concentration / km if km else 0

def cofactor_denominator(concentration, km):
    return 1 + concentration / km if km else 1

def construct_convenience_rate_law(stoichiometry, enzyme, cofactors_sets, partition, parameters):
    '''
    Make a convenience kinetics rate law for one enzyme

    Args:
        stoichiometry (dict): the stoichiometry for the given reaction
        enzyme (str): the current enzyme
        cofactors_sets: a list of lists with the required cofactors, grouped by [[cofactor set 1], [cofactor set 2]], each pair needs a kcat.
        partition: a list of lists. each sublist is the set of cofactors for a given partition.
            [[C1, C2],[C3, C4], [C5]]
        parameters (dict): all the parameters with {parameter_id: value}

    Returns:
        a kinetic rate law for the reaction, with arguments for concentrations and parameters,
        and returns flux.
    '''

    kcat_f = parameters.get('kcat_f')
    kcat_r = parameters.get('kcat_r')

    # remove km parameters with None as their value
    for parameter, value in parameters.items():
        if 'kcat' not in parameter:
            if value is None:
                for part in partition:
                    if parameter in part:
                        part.remove(parameter)
                for cofactors_set in cofactors_sets:
                    if parameter in cofactors_set:
                        cofactors_set.remove(parameter)
                # print('removing parameter: {}'.format(parameter))

	# if reversible, determine direction by looking at stoichiometry
    if kcat_r:
        coeff = [stoichiometry[mol] for mol in cofactors]
        positive_coeff = [c > 0 for c in coeff]
        if all(c == True for c in positive_coeff):  # if all coeffs are positive
            kcat = -kcat_r  # use reverse rate
        elif all(c == False for c in positive_coeff):  # if all coeffs are negative
            kcat = kcat_f
    else:
        kcat = kcat_f

    def rate_law(concentrations):

        # construct numerator
        enzyme_concentration = concentrations[enzyme]

        numerator = 0
        for cofactors in cofactors_sets:
            # multiply the affinities of all cofactors
            term = np.prod([
                cofactor_numerator(
                    concentrations[molecule],
                    parameters[molecule])  # km of molecule
                for molecule in cofactors])
            numerator += kcat * term  # TODO (if there is no kcat, need an exception)
        numerator *= enzyme_concentration

        # construct denominator, with all competing terms in the partition
        # denominator starts at +1 for the unbound state
        denominator = 1
        for cofactors_set in partition:
            # multiply the affinities of all cofactors in this partition
            term = np.prod([
                cofactor_denominator(
                    concentrations[molecule],
                    parameters[molecule])
				for molecule in cofactors_set])
            denominator += term - 1
        flux = numerator / denominator

        return flux

    return rate_law

# Make rate laws
def make_rate_laws(reactions, rate_law_configuration, kinetic_parameters):
    '''
    Make a rate law for each reaction

    Args:
        reactions (dict): in the same format as all_reactions, described above

        rate_law_configuration (dict): with an embedded structure:
            {enzyme_id: {
                'reaction_cofactors': {
                    reaction_id: [cofactors list]
                    }
                'partition': [partition list]
                }
            }

        kinetic_parameters (dict): with an embedded structure:
            {reaction_id: {
                'enzyme_id': {
                    parameter_id: value
                    }
                }
            }

    Returns:
        rate_laws (dict): each reaction_id is a key and has sub-dictionary for each relevant enzyme,
            with kinetic rate law functions as their values
    '''

    rate_laws = {reaction_id: {} for reaction_id in list(reactions.keys())}
    for reaction_id, specs in reactions.items():
        stoichiometry = specs.get('stoichiometry')
        # reversible = specs.get('is reversible') # TODO (eran) -- add reversibility based on specs
        enzymes = specs.get('catalyzed by')

        # rate law for each enzyme
        for enzyme in enzymes:
            if enzyme not in kinetic_parameters[reaction_id]:
                print('{} not in reaction {}'.format(enzyme, reaction_id))
                continue

            cofactors_sets = rate_law_configuration[enzyme]["reaction_cofactors"][reaction_id]
            partition = rate_law_configuration[enzyme]["partition"]

            rate_law = construct_convenience_rate_law(
                stoichiometry,
                enzyme,
                cofactors_sets,
                partition,
                kinetic_parameters[reaction_id][enzyme])

            # save the rate law for each enzyme in this reaction
            rate_laws[reaction_id][enzyme] = rate_law

    return rate_laws


class KineticFluxModel(object):
    '''
    A kinetic rate law class

    Args:
        all_reactions (dict): all metabolic reactions, with:
            {reaction_id: {
                'catalyzed by': list,
                'is reversible': bool,
                'stoichiometry': dict,
                }}

        kinetic_parameters (dict): a dictionary of parameters a nested format:
            {reaction_id: {
                enzyme_id : {
                    param_id: param_value}}}

    Attributes:
        rate_laws: a dict, with a key for each reaction id, and then subdictionaries with each reaction's enzymes
            and their rate law function. These rate laws are used directly from within this dictionary
    '''

    def __init__(self, all_reactions, kinetic_parameters):

        self.kinetic_parameters = kinetic_parameters
        self.reaction_ids = list(self.kinetic_parameters.keys())
        self.reactions = {reaction_id: all_reactions[reaction_id] for reaction_id in all_reactions}
        self.molecule_ids = get_molecules(self.reactions)

        # make the rate laws
        self.rate_law_configuration = make_configuration(self.reactions)

        self.rate_laws = make_rate_laws(
            self.reactions,
            self.rate_law_configuration,
            self.kinetic_parameters)

    def get_fluxes(self, concentrations_dict):
        '''
        Use rate law functions to calculate flux

        Args:
            concentrations_dict (dict): all relevant molecules and their concentrations, in mmol/L.
                {molecule_id: concentration}

        Returns:
            reaction_fluxes (dict) - with fluxes for all reactions
        '''

        # Initialize reaction_fluxes and exchange_fluxes dictionaries
        reaction_fluxes = {reaction_id: 0.0 for reaction_id in self.reaction_ids}

        for reaction_id, enzymes in self.rate_laws.items():
            for enzyme, rate_law in enzymes.items():
                flux = rate_law(concentrations_dict)
                reaction_fluxes[reaction_id] += flux

        return reaction_fluxes



toy_reactions = {
    'ABC-13-RXN': {
        'stoichiometry': {
            ('cytoplasm', 'PI'): 1,
            ('cytoplasm', 'ADP'): 1,
            ('cytoplasm', 'GLT'): 1,
            ('cytoplasm', 'PROTON'): 1,
            ('cytoplasm', 'ATP'): -1,
            ('cytoplasm', 'WATER'): -1,
            ('periplasm', 'GLT'): -1},
        'is reversible': False,
        'catalyzed by': [
            ('membrane', 'ABC-13-CPLX')]},
    'TRANS-RXN-122': {
        'stoichiometry': {
            ('cytoplasm', 'GLT'): 1, 
            ('periplasm', 'GLT'): -1, 
            ('periplasm', 'NA+'): -2, 
            ('cytoplasm', 'NA+'): 2},
        'is reversible': False,
        'catalyzed by': [
            ('membrane', 'GLTP-MONOMER'),
            ('membrane', 'DCTA-MONOMER')]},
    }


toy_kinetics = {
    'ABC-13-RXN': {
        ('membrane', 'ABC-13-CPLX'): {
            ('cytoplasm', 'ATP'): None,
            ('periplasm', 'GLT'): 1e-3,
            ('cytoplasm', 'WATER'): None,
            'kcat_f': 1.0
        }},
    'TRANS-RXN-122': {
        ('membrane', 'DCTA-MONOMER'): {
            ('periplasm', 'GLT'): 1e-3,
            ('periplasm', 'NA+'): 1e-5,
            'kcat_f': 1.0
        },
        ('membrane', 'GLTP-MONOMER'): {
            ('periplasm', 'GLT'): 1e-3,
            ('periplasm', 'NA+'): 1e-5,
            ('periplasm', 'PROTON'): None,
            'kcat_f': 1.0
        }}
    }

toy_initial_state = {
    'cytoplasm': {
        'PI': 1.0,
        'ADP': 1.0,
        'GLT': 1.0,
        'PROTON': 1.0,
        'ATP': 1.0,
        'WATER': 1.0,
        'NA+': 1.0},
    'periplasm': {
        'GLT': 1.0,
        'NA+': 1.0
    },
    'membrane': {
        'ABC-13-CPLX': 1.0,
        'GLTP-MONOMER': 1.0,
        'DCTA-MONOMER': 1.0}
}


def test_kinetics():
    kinetic_rate_laws = KineticFluxModel(toy_reactions, toy_kinetics)
    flattened_toy_states = tuplify_port_dicts(toy_initial_state)
    flux = kinetic_rate_laws.get_fluxes(flattened_toy_states)

    print(flux)


if __name__ == '__main__':
    test_kinetics()



import random
import copy
import numpy as np

from vivarium.library.datum import Datum
from vivarium_cell.data.chromosomes.toy_chromosome import toy_chromosome_config
from vivarium_cell.data.nucleotides import nucleotides
from vivarium_cell.library.polymerize import Polymerase, BindingSite, Terminator, Template, Elongation, polymerize_to, add_merge

INFINITY = float('inf')

def frequencies(l):
    '''
    Return number of times each item appears in the list.
    '''

    result = {}
    for item in l:
        if not item in result:
            result[item] = 0
        result[item] += 1
    return result

def rna_bases(sequence):
    sequence = sequence.replace('A', 'U')
    sequence = sequence.replace('T', 'A')
    sequence = sequence.replace('C', 'x')
    sequence = sequence.replace('G', 'C')
    sequence = sequence.replace('x', 'G')
    return sequence

def sequence_monomers(sequence, begin, end):
    if begin < end:
        subsequence = sequence[begin:end]
    else:
        subsequence = sequence[begin:end:-1]
    return subsequence

def traverse(tree, key, f, combine):
    '''
    Traverse the given tree starting using the `key` node as the root and calling `f` on each leaf,
    combining values with `combine` at each subsequent level to create new leaves for `f`.
    '''

    node = tree[key]
    if node.children:
        eldest = traverse(tree, node.children[0], f, combine)
        youngest = traverse(tree, node.children[1], f, combine)
        outcome = combine(eldest, youngest)
        return f(node, outcome)
    else:
        return f(node)


class Operon(Datum):
    defaults = {
        'id': '',
        'position': 0,
        'direction': 1,
        'length': 0,
        'genes': []}

    def __init__(self, config):
        super(Operon, self).__init__(config)

class Domain(Datum):
    defaults = {
        'id': 0,
        'lead': 0,
        'lag': 0,
        'children': []}

    def __init__(self, config):
        super(Domain, self).__init__(config)

    def contains(self, position):
        if position < 0:
            return position < self.lag
        else:
            return position > self.lead

    def surpassed(self, position, lead, lag):
        if position < 0:
            return position <= self.lag and position > self.lag + lag
        else:
            return position >= self.lead and position < self.lead + lead

    def random_child(self):
        return random.choice(self.children)

    def descendants(self, tree):
        return [self] + [tree[child].descendants(tree) for child in self.children]

class RnapTerminator(Terminator):
    def operon_from(self, genes, promoter):
        return Operon({
            'id': self.products[0], # assume there is only one operon product
            'position': promoter.position,
            'direction': promoter.direction,
            'length': (self.position - promoter.position) * promoter.direction,
            'genes': genes.get(self.products[0], [])})

class Promoter(Template):
    '''
    Promoters are the main unit of expression. They define a direction of
    polymerization, contain binding sites for transcription factors, and declare some
    number of terminators, each of which has a strength and corresponds to a particular
    operon if chosen.
    '''

    schema = {
        'sites': BindingSite,
        'terminators': RnapTerminator}

    def operons(self, genes):
        return [
            terminator.operon_from(genes, self)
            for terminator in self.terminators]

class Rnap(Polymerase):
    defaults = {
        'id': 0,
        'template': None,
        'template_index': 0,
        'terminator': 0,
        'domain': 0,
        'state': None, # other states: ['bound', 'polymerizing', 'occluding', 'complete']
        'position': 0}

    def __init__(self, config):
        super(Rnap, self).__init__(config)

class Chromosome(Datum):
    schema = {
        'promoters': Promoter,
        'domains': Domain,
        'rnaps': Rnap}

    defaults = {
        'sequence': '',
        'genes': {},
        'promoters': {},
        'promoter_order': [],
        'domains': {
            0: {
                'id': 0,
                'lead': 0,
                'lag': 0,
                'children': []}},
        'root_domain': 0,
        'rnap_id': 0,
        'rnaps': {}}

    def operons(self):
        return [
            operon
            for promoter in self.promoters.values()
            for operon in promoter.operons(self.genes)]

    def copy_number(self, position, domain_key=None):
        if not domain_key:
            domain_key = self.root_domain
        domain = self.domains[domain_key]
        if domain.contains(position):
            return 1
        else:
            return sum([
                self.copy_number(position, child)
                for child in domain.children])

    def promoter_copy_numbers(self):
        copy_numbers = [
            self.copy_number(self.promoters[promoter_key].position)
            for promoter_key in self.promoter_order]
        return np.array(copy_numbers)

    def promoter_rnaps(self):
        by_promoter = {
            promoter_key: {}
            for promoter_key in self.promoter_order}

        for rnap in self.rnaps.values():
            if rnap.is_occluding():
                by_promoter[rnap.template][rnap.domain] = rnap

        return by_promoter

    def promoter_domains(self):
        return {
            promoter_key: self.position_domains(
                self.root_domain,
                self.promoters[promoter_key].position)
            for promoter_key in self.promoter_order}

    def position_domains(self, domain_index, position):
        domain = self.domains[domain_index]
        if len(domain.children) == 0 or (position < 0 and domain.lag >= position) or (position >= 0 and domain.lead <= position):
            return set([domain_index])
        else:
            return set.union(*[
                self.position_domains(child, position)
                for child in domain.children])

    def apply_thresholds(self, thresholds):
        for path, level in thresholds.items():
            promoter, factor = path
            found = False
            for site in self.promoters[promoter].sites:
                if factor in site.thresholds:
                    site.thresholds[factor] = level
                    found = True
            if not found:
                print('binding site not found for {} with level {}'.format(path, level))

    def bind_rnap(self, promoter_index, domain):
        self.rnap_id += 1
        promoter_key = self.promoter_order[promoter_index]

        new_rnap = Rnap({
            'id': self.rnap_id,
            'template': promoter_key,
            'template_index': promoter_index,
            'domain': domain,
            'position': 0})
        new_rnap.bind()
        self.rnaps[new_rnap.id] = new_rnap
        return new_rnap

    def terminator_distance(self):
        distance = INFINITY
        for rnap in self.rnaps.values():
            if rnap.is_polymerizing():
                promoter = self.promoters[rnap.template]

                # rnap position is relative to the promoter it is bound to
                rnap_position = promoter.absolute_position(rnap.position)
                terminator_index = promoter.next_terminator(rnap_position)
                rnap.terminator = terminator_index
                terminator = promoter.terminators[terminator_index]
                span = abs(terminator.position - rnap_position)
                if span < distance:
                    distance = span

        if distance == INFINITY:
            distance = 1
        return distance

    def sequences(self):
        return {
            promoter_key: rna_bases(sequence_monomers(
                self.sequence,
                promoter.position,
                promoter.last_terminator().position))
            for promoter_key, promoter in self.promoters.items()}

    def product_sequences(self):
        sequences = {}
        for promoter_key, promoter in self.promoters.items():
            for terminator in promoter.terminators:
                for product in terminator.products:
                    sequences[product] = rna_bases(sequence_monomers(
                        self.sequence,
                        promoter.position,
                        terminator.position))

        return sequences

    def next_polymerize(self, elongation_limit=INFINITY, monomer_limits={}):
        distance = self.terminator_distance()
        elongate_to = min(elongation_limit, distance)

        sequences = self.sequences()

        monomers, monomer_limits, terminated, complete_transcripts, self.rnaps = polymerize_to(
            sequences,
            self.rnaps,
            self.promoters,
            elongate_to,
            nucleotides,
            monomer_limits)

        return elongate_to, monomers, monomer_limits, complete_transcripts

    def polymerize(self, elongation, monomer_limits):
        iterations = 0
        attained = 0
        all_monomers = {}
        complete_transcripts = {}

        while attained < elongation:
            elongated, monomers, monomer_limits, complete = self.next_polymerize(
                elongation_limit=elongation - attained,
                monomer_limits=monomer_limits)
            attained += elongated
            all_monomers = add_merge([all_monomers, monomers])
            complete_transcripts = add_merge([complete_transcripts, complete])
            iterations += 1

        return iterations, all_monomers, complete_transcripts, monomer_limits

    def initiate_replication(self):
        leaves = [leaf for leaf in self.domains.values() if not leaf.children]
        next_id = max([leaf.id for leaf in leaves]) + 1
        for leaf in leaves:
            for child in [0, 1]:
                domain = Domain({'id': next_id + child})
                self.domains[domain.id] = domain
            leaf.children = [next_id, next_id + 1]
            next_id += 2

    def advance_replisomes(self, distances):
        '''
        distances is a dictionary of domain ids to tuples of how far each strand advances
        of the form (lead, lag)
        '''
        for domain_key, distance in distances.items():
            domain = self.domains[domain_key]
            lead, lag = distances[domain_key]

            for rnap in self.rnaps.values():
                if rnap.domain == domain_key:
                    promoter = self.promoters[rnap.template]
                    position = promoter.position
                    position += rnap.position * promoter.direction
                    if domain.surpassed(position, lead, -lag):
                        rnap.domain = domain.random_child()

            domain.lead += lead
            domain.lag -= lag

    def divide_chromosome(self, domain, division=None):
        if not division:
            division = {
                'sequence': self.sequence,
                'promoters': {id: promoter.to_dict() for id, promoter in self.promoters.items()},
                'domains': {domain.id: domain.to_dict()},
                'root_domain': domain.id,
                'rnaps': {
                    rnap.id: rnap.to_dict()
                    for rnap in self.rnaps.items()
                    if rnap.domain == domain.id}}

        else:
            division['domains'][domain.id] = domain.to_dict()
            for rnap in self.rnaps.values():
                if rnap.domain == domain.id:
                    division['rnaps'].append(rnap.to_dict())

        return division

    def combine_state(self, a, b):
        merge = copy.deepcopy(a)
        merge['domains'].update(b['domains'])
        merge['rnaps'].extend(b['rnaps'])
        return merge

    def terminate_replication(self):
        children = self.domains[self.root_domain].children
        divided = [
            traverse(
                self.domains,
                child,
                self.divide_chromosome,
                self.combine_state)
            for child in children]

        return [Chromosome(fork) for fork in divided]

    def __init__(self, config):
        super(Chromosome, self).__init__(config)
        if self.promoter_order:
            self.promoter_order = list(self.promoters.keys())



def test_chromosome():
    chromosome = Chromosome(toy_chromosome_config)
    print(chromosome.promoters['pA'].terminators[0].products)
    print(chromosome)

    print('operons:')
    print(chromosome.operons())

    chromosome.initiate_replication()
    print(chromosome.domains)
    assert len(chromosome.domains) == 3

    chromosome.advance_replisomes({0: (5, 7)})
    print('replisomes:')
    print(chromosome)

    # chromosome.initiate_replication()
    # print(chromosome.domains)
    # assert len(chromosome.domains) == 7

    # chromosome.advance_replisomes({0: (7, 5), 1: (3, 4), 2: (8, 9)})
    # print('replisomes:')
    # print(chromosome)

    print('copy numbers 1 4 7 -9 11')
    print(chromosome.copy_number(1))
    print(chromosome.copy_number(4))
    print(chromosome.copy_number(7))
    print(chromosome.copy_number(-9))
    print(chromosome.copy_number(11))

    print('promoter copy numbers')
    print(chromosome.promoter_copy_numbers())

    print('promoter rnaps')
    print(chromosome.promoter_rnaps())

    print('promoter domains')
    print(chromosome.promoter_domains())

    print('rnaps')
    print([rnap.to_dict() for rnap in chromosome.rnaps.values()])

    print('completed after advancing 5')
    print(chromosome.polymerize(5, {
        'rATP': 100, 'rUTP': 100, 'rGTP': 100, 'rCTP': 100}))

    print('rnaps after polymerizing')
    print(chromosome.rnaps)
    
    children = chromosome.terminate_replication()
    print('termination:')
    print(children)

    return chromosome

if __name__ == '__main__':
    test_chromosome()



def invert_index(mapping):
    inverse = {}
    for canon, synonyms in mapping.items():
        for synonym in synonyms:
            inverse[synonym] = canon
    return inverse

synonyms = {
    # 'ATP': ['atp_c'],
    'ADP': ['adp_c'],
    'GTP': ['gtp_c'],
    'NAD(H)': ['nad_c'],
    'NADP(H)': ['nadp_c'],
    'FAD(H2)': ['fad_c'],
    'ATP': ['A', 'ATP', 'atp_c'],  # ATP name conflicts with the energy carrier
    'GTP': ['G', 'GTP', 'gtp_c'],
    'UTP': ['U', 'UTP', 'utp_c'],
    'CTP': ['C', 'CTP', 'ctp_c'],
    'Alanine': ['ala__L_c'],
    'Arginine': ['arg__L_c'],
    'Asparagine': ['asn__L_c'],
    'Aspartate': ['asp__L_c'],
    'Cysteine': ['cys__L_c'],
    'Glutamate': ['glu__L_c'],
    'Glutamine': ['gln__L_c'],
    'Glycine': ['gly_c'],
    'Histidine': ['his__L_c'],
    'Isoleucine': ['ile__L_c'],
    'Leucine': ['leu__L_c'],
    'Lysine': ['lys__L_c'],
    'Methionine': ['met__L_c'],
    'Phenylalanine': ['phe__L_c'],
    'Proline': ['pro__L_c'],
    'Serine': ['ser__L_c'],
    'Threonine': ['thr__L_c'],
    'Tryptophan': ['trp__L_c'],
    'Tyrosine': ['tyr__L_c'],
    'Valine': ['val__L_c'],
}

index = invert_index(synonyms)

def get_synonym(id):
    return index.get(id, id)




'''
Execute by running: ``python vivarium/process/template_process.py``
Note: This process is unfinished.
'''

import os
import numpy as np

from vivarium.core.process import Deriver
from vivarium.core.composition import (
    simulate_process_in_experiment,
    PROCESS_OUT_DIR,
)
from vivarium.plots.simulation_output import plot_simulation_output

'''
References:
 - width of 0.73 um: Cluzel et al., Nucleic Acids Research (2008)
	- strain: K-12 Frag1
 - periplasm width of 21 nm: Matias et al., J Bacteriology (2003)
 - cell density of 1100 g/L (from WCM): Baldwin et al., Arch microbiol. (1995)
'''

NAME = 'spatial_geometry'
WIDTH = 0.73		# in um
PERIPLASM_WIDTH = 21        # in nm
DENSITY = 1100      # fg/um^3

class SpatialGeometry(Deriver):
    """
    Calculates geometric parameters of nodes and edges utilized in
    diffusion_network.py. This process assumes E. coli is a spherocylinder
    with constant radius. This process is currently unfinished. This process
    assumes that there are properties of the nodes that specify which type
    of compartment it is such that the geometry rules below apply. This is
    not currently implemented in ecoli_spatial.py in vivarium-ecoli.
    Much of the geometry rules have been specified, however it should be
    checked to see if it leads to consistent results.
    """

    name = NAME
    defaults = {
        'nodes': [
            'nucleoid',
            'periplasm',
        ],
        'edges': {},
        'density': DENSITY,
        'mw': {},
        'width': WIDTH,
    }

    def __init__(self, parameters=None):
        super(SpatialGeometry, self).__init__(parameters)
        self.nodes = np.asarray(self.parameters['nodes'])
        self.edges = self.parameters['edges']
        self.mw = self.parameters['mw']
        self.molecule_ids = self.parameters['mw'].keys()
        self.density = self.parameters['density']

    def initial_state(self, config=None):
        pass

    def ports_schema(self):
        node_schema = {
            node_id: {
                'volume': {
                    '_default': 1.0,
                },
                'length': {
                    '_default': 1.0,
                },
                'radius': {
                    '_default': 1.0,
                },
                'molecules': {
                    '*': {
                        '_default': 0,
                    }
                }
            } for node_id in self.parameters['nodes']
        }
        edge_schema = {
            edge_id: {
                'nodes': [],
                'cross_sectional_area': 1.0,
            } for edge_id in self.parameters['edges'].keys()
        }

        return {
            'nodes': node_schema,
            'edges': edge_schema
        }

    def next_update(self, timestep, states):
        volume = np.zeros(len(self.nodes))
        nodes = states['nodes']
        length = [nodes[node_id]['length'] for node_id in self.nodes]
        cross_sectional_area = np.zeros(len(self.edges))
        inner_radius = nodes['nucleoid']['radius']
        outer_radius = nodes['periplasm']['radius'] + inner_radius

        # TODO -- get volume of each node from molecules and density
        for i, node_id in enumerate(self.nodes):
            # if nucleoid or nucleoid region of membrane
            nucleoid = True
            volume[i] = sum(array_from(nodes[node_id]['molecules']))*self.density
            if nucleoid:
                length[i] = volume[i]/(np.pi * inner_radius**2)
        for i, edge_id in enumerate(self.edges):
            nodes = states[edge_id]['nodes']
            nucleoid = 'nucleoid' in nodes
            cytosol = 'cytosol_front' or 'cytosol_rear' in nodes
            periplasm = 'periplasm' in nodes
            cytosol_inner_membrane = False
            cytosol_outer_membrane = False
            nucleoid_inner_membrane = False
            nucleoid_outer_membrane = False

            # series of if statements for each interface
            if nucleoid and cytosol:
                cross_sectional_area[i] = np.pi * inner_radius**2
                continue
                # find if node is membrane
            if nucleoid and nucleoid_inner_membrane:
                node_id = 'membrane'
                cross_sectional_area[i] = 2 * np.pi * inner_radius * states[
                    'nodes'][node_id]['length']
                continue
            if cytosol and cytosol_inner_membrane:
                cross_sectional_area[i] = 2 * np.pi * inner_radius**2
                continue
            if periplasm and nucleoid_inner_membrane:
                cross_sectional_area[i] = 2 * np.pi * inner_radius * states[
                        'nodes'][node_id]['length']
                continue
            if periplasm and nucleoid_outer_membrane:
                cross_sectional_area[i] = 2 * np.pi * outer_radius * states[
                    'nodes'][node_id]['length']
                continue
            if periplasm and cytosol_inner_membrane:
                cross_sectional_area[i] = 2 * np.pi * inner_radius**2
                continue
            if periplasm and cytosol_outer_membrane:
                cross_sectional_area[i] = 2 * np.pi * outer_radius**2
                continue

        node_update = {
            node_id: {
                'volume': volume[np.where(self.nodes == node_id)][0],
                'length': volume[np.where(self.nodes == node_id)][0],
            } for node_id in self.nodes
        }
        edge_update = {}
        # edge_update = {
        #     edge_id: {
        #         'cross_sectional_area': cross_sectional_area[
        #             np.where(self.edges == edge_id)][0],
        #     } for edge_id in self.nodes
        # }
        return {**node_update, **edge_update}





# functions to configure and run the process
def run_spatial_geometry_process():
    '''Run a simulation of the process.

    Returns:
        The simulation output.
    '''

    # initialize the process by passing in parameters
    parameters = {}
    spatial_geometry_process = SpatialGeometry(parameters)

    # declare the initial state, mirroring the ports structure
    initial_state = {}

    # run the simulation
    sim_settings = {
        'total_time': 10,
        'initial_state': initial_state}
    output = simulate_process_in_experiment(spatial_geometry_process, sim_settings)

    return output


def test_spatial_geometry_process():
    '''Test that the process runs correctly.

    This will be executed by pytest.
    '''
    output = run_spatial_geometry_process()
    # TODO: Add assert statements to ensure correct performance.

    return output


def main():
    '''Simulate the process and plot results.'''
    # make an output directory to save plots
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    output = test_spatial_geometry_process()

    # # plot the simulation output
    # plot_settings = {}
    # plot_simulation_output(output, plot_settings, out_dir)


# Helper functions
def array_from(d):
    return np.array(list(d.values()))


def array_to(keys, array):
    return {
        key: array[index]
        for index, key in enumerate(keys)}

# run module with python vivarium/process/template_process.py
if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function

import os
import uuid
import math
import random

from vivarium.core.experiment import (
    generate_state,
    Experiment
)
from vivarium.core.composition import (
    make_agents,
    simulate_experiment,
)
from vivarium.plots.agents_multigen import plot_agents_multigen

# composites
from vivarium_cell.composites.lattice import Lattice
from vivarium_cell.composites.growth_division_minimal import GrowthDivisionMinimal

# processes
from vivarium.library.units import units
from vivarium_cell.processes.derive_globals import volume_from_length
from vivarium_cell.processes.multibody_physics import (
    volume_from_length, DEFAULT_BOUNDS, PI
)
from vivarium_cell.plots.multibody_physics import plot_snapshots


def mother_machine_experiment(config):
    # configure the experiment
    agent_ids = config.get('agent_ids', [])
    emitter = config.get('emitter', {'type': 'timeseries'})

    # get the environment
    environment = Lattice(config.get('environment', {}))
    network = environment.generate({})
    processes = network['processes']
    topology = network['topology']

    # get the agents
    growth_division = GrowthDivisionMinimal(config.get('growth_division', {}))
    agents = make_agents(agent_ids, growth_division, config.get('growth_division', {}))
    processes['agents'] = agents['processes']
    topology['agents'] = agents['topology']

    return Experiment({
        'processes': processes,
        'topology': topology,
        'initial_state': config.get('initial_state', {}),
        'emitter': emitter,
    })



# configurations
def mother_machine_body_config(config):
    '''
    Gets initial agent body locations given the mother machine set up
    '''
    # cell dimensions
    width = 1
    length = 2
    volume = volume_from_length(length, width)

    agent_ids = config['agent_ids']
    bounds = config.get('bounds', DEFAULT_BOUNDS)
    channel_space = config.get('channel_space', 1)
    n_agents = len(agent_ids)

    # possible locations, shuffled for index-in
    n_spaces = math.floor(bounds[0]/channel_space)
    assert n_agents < n_spaces, 'more agents than mother machine spaces'

    possible_locations = [
        [x*channel_space - channel_space/2, 0.01]
        for x in range(1, n_spaces)]
    random.shuffle(possible_locations)

    initial_agents = {
        agent_id: {
            'boundary': {
                'location': possible_locations[index],
                'angle': PI/2,
                'volume': volume,
                'length': length,
                'width': width}}
        for index, agent_id in enumerate(agent_ids)}
    return initial_agents


def get_mother_machine_config():
    bounds = [20, 20]
    n_bins = [10, 10]
    channel_height = 0.7 * bounds[1]
    channel_space = 1.5
    n_agents = 3

    agent_ids = [str(agent_id) for agent_id in range(n_agents)]

    ## growth division agent
    growth_division_config = {
        'agents_path': ('..', '..', 'agents'),
        'global_path': ('global',),
        'growth_rate': 0.006,  # fast!
        'division_volume': 2.6}

    ## environment
    # multibody
    multibody_config = {
        'animate': False,
        'mother_machine': {
            'channel_height': channel_height,
            'channel_space': channel_space},
        'jitter_force': 1e-3,
        'bounds': bounds}

    body_config = {
        'bounds': bounds,
        'channel_height': channel_height,
        'channel_space': channel_space,
        'agent_ids': agent_ids}
    initial_agents = mother_machine_body_config(body_config)

    # diffusion
    diffusion_config = {
        'molecules': ['glc'],
        'gradient': {
            'type': 'gaussian',
            'molecules': {
                'glc':{
                    'center': [0.5, 0.5],
                    'deviation': 3},
            }},
        'diffusion': 5e-3,
        'n_bins': n_bins,
        'size': bounds}

    return {
        'initial_state': {
            'agents': initial_agents},
        'agent_ids': agent_ids,
        'growth_division': growth_division_config,
        'environment': {
            'multibody': multibody_config,
            'diffusion': diffusion_config}}


def run_mother_machine(time=5, out_dir='out'):
    mother_machine_config = get_mother_machine_config()
    experiment = mother_machine_experiment(mother_machine_config)

    # simulate
    settings = {
        'emit_step': 5,
        'total_time': time,
        'return_raw_data': True}
    data = simulate_experiment(experiment, settings)

    # agents plot
    plot_settings = {
        'agents_key': 'agents'}
    plot_agents_multigen(data, plot_settings, out_dir)

    # snapshot plot
    multibody_config = mother_machine_config['environment']['multibody']
    agents = {time: time_data['agents'] for time, time_data in data.items()}
    fields = {time: time_data['fields'] for time, time_data in data.items()}
    snapshot_data = {
        'agents': agents,
        'fields': fields,
        'config': multibody_config}
    plot_config = {
        'out_dir': out_dir,
        'filename': 'snapshots'}

    plot_snapshots(snapshot_data, plot_config)


if __name__ == '__main__':
    out_dir = os.path.join('out', 'experiments', 'mother_machine')
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    run_mother_machine(400, out_dir)



import numpy as np

from vivarium.core.process import Deriver
from vivarium.library.units import units, remove_units

from vivarium_cell.library.lattice_utils import (
    get_bin_site,
    get_bin_volume,
    count_to_concentration,
)

CONCENTRATION_UNIT = 1  # TODO (ERAN) set value -- units.ng / units.mL
LENGTH_UNIT = units.um


class LocalField(Deriver):

    name = 'local_field'
    defaults = {}

    def __init__(self, parameters=None):
        super(LocalField, self).__init__(parameters)

    def ports_schema(self):
         return {
            'exchanges': {
                '*': {
                    '_default': 0,  # counts!
                }
            },
            'location': {
                '_default': [0.5 * LENGTH_UNIT, 0.5 * LENGTH_UNIT]
            },
            'fields': {
                '*': {
                    '_default': np.ones(1),
                    '_updater': 'accumulate',
                }
            },
            'dimensions': {
                'bounds': {
                    '_default': [1, 1],
                },
                'n_bins': {
                    '_default': [1, 1],
                },
                'depth': {
                    '_default': 1,
                },
            }
        }


    def next_update(self, timestep, states):
        location = remove_units(states['location'])
        n_bins = states['dimensions']['n_bins']
        bounds = states['dimensions']['bounds']
        depth = states['dimensions']['depth']
        exchanges = states['exchanges']

        # get bin
        bin_site = get_bin_site(location, n_bins, bounds)
        bin_volume = get_bin_volume(n_bins, bounds, depth) * units.L

        # apply exchanges
        delta_fields = {}
        reset_exchanges = {}
        for mol_id, value in exchanges.items():
            delta_fields[mol_id] = np.zeros(
                (n_bins[0], n_bins[1]), dtype=np.float64)
            exchange = value * units.count
            concentration = count_to_concentration(exchange, bin_volume)
            delta_fields[mol_id][bin_site[0], bin_site[1]] += concentration.to(
                units.mmol / units.L).magnitude
            reset_exchanges[mol_id] = {
                '_value': 0,
                '_updater': 'set'}

        return {
            'exchanges': reset_exchanges,
            'fields': delta_fields}


def test_local_fields():
    parameters = {}
    local_fields_process = LocalField(parameters)

    bounds = [5, 5]
    n_bins = [3, 3]
    initial_state = {
        'exchanges': {
            'A': 20
        },
        'location': [0.5, 0.5],
        'fields': {
            'A': np.ones((n_bins[0], n_bins[1]), dtype=np.float64)
        },
        'dimensions': {
            'bounds': bounds,
            'n_bins': n_bins,
            'depth': 1,
        }
    }

    output = local_fields_process.next_update(0, initial_state)


if __name__ == '__main__':
    test_local_fields()



from __future__ import absolute_import, division, print_function

import numpy as np
from scipy import constants

from vivarium.library.dict_utils import deep_merge
from vivarium.library.units import units, remove_units, Quantity
from vivarium.core.process import Process


AVOGADRO = constants.N_A * 1 / units.mol


class CellEnvironmentDiffusion(Process):

    name = "cell_environment_diffusion"
    defaults = {
        'molecules_to_diffuse': [],
        'default_state': {
            'global': {
                'volume': 1.2 * units.fL,
            }
        },
        'default_default': 0,
        'permeabilities': {
            'porin': 1e-1,
        },
    }

    def ports_schema(self):

        schema = {
            'internal': {
                # Molecule concentration in mmol/L
                molecule: {
                    '_default': self.parameters['default_default'],
                    '_divider': 'set',
                }
                for molecule in self.parameters['molecules_to_diffuse']
            },
            'external': {
                # Molecule concentration in mmol/L
                molecule: {
                    '_default': self.parameters['default_default'],
                    '_divider': 'set',
                }
                for molecule in self.parameters['molecules_to_diffuse']
            },
            'membrane': {
                # Porin concentration in mmol/L
                porin: {
                    '_default': self.parameters['default_default'],
                    '_emit': True,
                }
                for porin in self.parameters['permeabilities']
            },
            'fields': {
                # Molecule count in mmol
                molecule: {
                    '_default': np.full(
                        (1, 1), self.parameters['default_default']),
                }
                for molecule in self.parameters['molecules_to_diffuse']
            },
            'global': {
                'volume': {
                    '_default': self.parameters['default_default'],
                },
                'location': {
                    '_default': [0.5, 0.5],
                },
            },
            'dimensions': {
                'bounds': {
                    '_default': [1, 1],
                },
                'n_bins': {
                    '_default': [1, 1],
                },
                'depth': {
                    '_default': 1,
                },
            },
        }

        for port, port_conf in self.parameters['default_state'].items():
            for variable, default in port_conf.items():
                schema[port][variable]['_default'] = default

        return schema

    def next_update(self, timestep, states):
        permeabilities = self.parameters['permeabilities']
        rates = {
            molecule: sum([
                states['membrane'][porin] * permeability
                for porin, permeability in permeabilities.items()
            ])
            for molecule in self.parameters['molecules_to_diffuse']
        }
        # Flux is positive when leaving the cell
        flux_mmol = {
            molecule: (
                states['internal'][molecule]
                - states['external'][molecule]
            ) * rate * timestep * units.mmol
            for molecule, rate in rates.items()
        }
        flux_counts = {
            molecule: flux * AVOGADRO
            for molecule, flux in flux_mmol.items()
        }
        if not isinstance(states['global']['volume'], Quantity):
            cell_volume = states['global']['volume'] * units.fL
        else:
            cell_volume = states['global']['volume']
        update = {
            'fields': {
                molecule: mol_flux.to(units.count).magnitude
                for molecule, mol_flux in flux_counts.items()
            },
            'internal': {
                molecule: - (
                    mol_flux / cell_volume
                ).to(units.mmol / units.L).magnitude
                for molecule, mol_flux in flux_mmol.items()
            },
        }
        return update



import os
import uuid
import copy

from vivarium.library.units import units
from vivarium.core.process import Composite
from vivarium.core.composition import (
    COMPARTMENT_OUT_DIR,
    simulate_compartment_in_experiment,
)
from vivarium.plots.agents_multigen import plot_agents_multigen

# processes
from vivarium.processes.meta_division import MetaDivision
from vivarium.processes.tree_mass import TreeMass
from vivarium_cell.processes.growth_protein import GrowthProtein
from vivarium_cell.processes.minimal_expression import (
    MinimalExpression,
    get_toy_expression_config,
)
from vivarium_cell.processes.convenience_kinetics import (
    ConvenienceKinetics,
    get_glc_lct_config
)


NAME = 'growth_division'


class GrowthDivision(Composite):

    defaults = {
        'boundary_path': ('boundary',),
        'agents_path': ('..', '..', 'agents',),
        'transport': get_glc_lct_config(),
        'daughter_path': tuple(),
        'fields_path': ('fields',),
        'dimensions_path': ('dimensions',),
        'growth': {},
        'expression': get_toy_expression_config(),
        'mass': {},
    }

    def __init__(self, config):
        super(GrowthDivision, self).__init__(config)

        # transport configs
        boundary_path = self.config['boundary_path']
        self.config['transport'] = self.config['transport']
        self.config['transport']['global_deriver_config'] = {
            'type': 'globals',
            'source_port': 'global',
            'derived_port': 'global',
            'global_port': boundary_path,
            'keys': []}

    def generate_processes(self, config):
        daughter_path = config['daughter_path']
        agent_id = config['agent_id']

        growth = GrowthProtein(config['growth'])
        transport = ConvenienceKinetics(config['transport'])
        expression = MinimalExpression(config['expression'])
        mass_deriver = TreeMass(config['mass'])

        # configure division
        division_config = dict(
            config.get('division', {}),
            daughter_path=daughter_path,
            agent_id=agent_id,
            generator=self)
        division = MetaDivision(division_config)

        return {
            'transport': transport,
            'growth': growth,
            'expression': expression,
            'mass_deriver': mass_deriver,
            'division': division,
        }

    def generate_topology(self, config):
        boundary_path = config['boundary_path']
        agents_path = config['agents_path']
        external_path = boundary_path + ('external',)
        fields_path = config['fields_path']
        dimensions_path = config['dimensions_path']
        return {
            'transport': {
                'internal': ('internal',),
                'external': external_path,
                'fields': fields_path,
                'fluxes': ('fluxes',),
                'global': boundary_path,
                'dimensions': dimensions_path,
            },

            'growth': {
                'internal': ('internal',),
                'global': boundary_path
            },

            'mass_deriver': {
                'global': boundary_path
            },

            'division': {
                'global': boundary_path,
                'agents': agents_path
            },

            'expression': {
                'internal': ('internal',),
                'external': external_path,
                'concentrations': ('internal_concentrations',),
                'global': boundary_path
            },
        }



if __name__ == '__main__':
    out_dir = os.path.join(COMPARTMENT_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    agent_id = '0'
    compartment = GrowthDivision({'agent_id': agent_id})

    # settings for simulation and plot
    settings = {
        'environment': {
            'volume': 1e-6 * units.L,  # L
            'ports': {
                'fields': ('fields',),
                'external': ('boundary', 'external',),
                'global': ('boundary',),
                'dimensions': ('dimensions',),
            },
        },
        'outer_path': ('agents', agent_id),  # TODO -- need to set the agent_id through here?
        'return_raw_data': True,
        'timestep': 1,
        'total_time': 500}
    output_data = simulate_compartment_in_experiment(compartment, settings)

    plot_settings = {}
    plot_agents_multigen(output_data, plot_settings, out_dir)



import os
import numpy as np
import matplotlib.pyplot as plt
import scipy.constants

from iteround import saferound

from vivarium.core.process import Process
from vivarium.core.composition import (
    simulate_process_in_experiment,
    PROCESS_OUT_DIR,
)

NAME = 'diffusion_network'

class DiffusionNetwork(Process):
    ''' Models Brownian diffusion using a network of nodes and edges. Each
        node acts as a cellular compartment and each edge connects those
        compartments, indicating where diffusion can occur.

         This :term:`process class` models diffusion based off of Fick's law.
         The following equation is used:

         * Diffusion: :math:`\\frac{dc}{dt} = \\frac{DA}{V} * \\frac{dc}{dx}`

             * :math:`D`: Diffusion constant
             * :math:`A`: Cross-sectional area of edge
             * :math:`V`: Volume of node

         This diffusion equation is solved using implicit Euler to derive a
         matrix, M, that updates the concentration, c(t), on each timestep.

         * Concentration update: :math:`\\c^{t+1} = M^{-1} * c^{t}`

         This process takes in molecular weights in order to solve for
         molecule hydrodynamic radii, which is it turn used to solve for
         diffusion constants. These calculations assume that all molecules are
         spherical proteins. For more information on how this is done, see
         `calculate_rp_from_mw` and `compute_diffusion_constants_from_rp`.
         However, if a molecule radius is known, it can be passed in as
         `radii`. Similarly, if the diffusion constant is known, it can be
         passed in as a property of an edge as `diffusion_constants`.


         .. note::
             This model treats all molecule classes as concentrations and is
             a deterministic solution. This model should only be used when there
             is a sufficient number of molecules such that they can be treated
             deterministically.

         :term:`Ports`:

         * **nodes**: Expects a :term:`store` which is a dict of node names
         (the keys of the dict) to a dict, which has the key value pairs
         for `length`, `volume`, and `molecules`.

         Arguments:
             initial_parameters: A dictionary of configuration options.
                 The following configuration options may be provided:

                 * **nodes** (:py:class:`list`): A list of node names.
                 * **edges** (:py:class:`dict`): Maps edge
                   names (the keys of the dict) to a dict (the
                   values of the dict), which must include the key-value
                   pairs of `nodes` to a list of nodes each edge connects,
                   and `cross_sectional_area` to the area of that edge.
                   Additionally, known diffusion constants can be included as
                   `diffusion_constants` in units of um^2/s, and edge-specific
                   scaling of the diffusion constants can be included as,
                   `diffusion_scaling_constant`.
                 * **mw** (:py:class:`dict`): Maps from
                   names of molecules (the keys of the dict) to their
                   molecular weights in units of fg (the values of the dict).
                 * **mesh_size** (:py:class:`float`): Mesh size in units of nm.
                 * **time_step** (:py:class:`float`): The time step used in
                   units of s.
                 * **radii** (:py:class:`dict`): Maps from molecule names
                   of molecules (the keys of the dict) to their known
                   hydrodynamic radii in units of nm (the values of the dict).
                   This is an optional parameter.
                 * **temp** (:py:class:`float`): Temperature of experiment
                 in units of K. This is an optional parameter.

'''

    name = NAME

    defaults = {
        'nodes': [],
        'edges': {},
        'mw': {},   # in fg
        'mesh_size': 50,    # in nm
        'time_step': 0.1,     # in s
        'radii': {},        # nm
        'temp': 310.15,     # in K
    }

    def __init__(self, parameters=None):
        super(DiffusionNetwork, self).__init__(parameters)
        self.nodes = np.asarray(self.parameters['nodes'])
        self.edges = self.parameters['edges']
        self.mw = self.parameters['mw']
        self.molecule_ids = self.parameters['mw'].keys()
        self.mesh_size = self.parameters['mesh_size']
        self.radii = self.parameters['radii']
        self.temp = self.parameters['temp']
        self.remainder = np.zeros((len(self.nodes), len(self.molecule_ids)))

        # get molecule radii by molecular weights
        self.rp = calculate_rp_from_mw(self.molecule_ids, self.mw)
        for mol_id, r in self.radii.items():
            self.rp[np.where(np.asarray(list(self.molecule_ids)) == mol_id)[0][0]] = r

        # get diffusion constants per molecule
        self.diffusion_constants = compute_diffusion_constants_from_rp(
            self.molecule_ids, self.rp, self.mesh_size, self.edges, self.temp)

    def ports_schema(self):
        '''
        ports_schema returns a dictionary that declares how each state will behave.
        Each key can be assigned settings for the schema_keys declared in Store:

        * `_default`
        * `_updater`
        * `_divider`
        * `_value`
        * `_properties`
        * `_emit`
        * `_serializer`
        '''
        schema = {
            node_id: {
                'volume': {
                    '_default': 1.0,
                },
                'length': {
                    '_default': 1.0,
                },
                'molecules': {
                    '*': {
                        '_default': 0,
                        '_emit': True,
                        '_updater': 'accumulate',
                    }
                },
            } for node_id in self.parameters['nodes']
        }
        return schema

    def next_update(self, timestep, state):
        M = np.asarray([np.identity(len(self.nodes)) for mol in self.molecule_ids])

        # construct M matrix based off of graph, all edges assumed bidirectional
        for edge_id, edge in self.edges.items():
            node_index_1 = np.where(self.nodes == edge['nodes'][0])[0][0]
            node_index_2 = np.where(self.nodes == edge['nodes'][1])[0][0]
            cross_sectional_area = edge['cross_sectional_area']
            vol_1 = state[edge['nodes'][0]]['volume']
            vol_2 = state[edge['nodes'][1]]['volume']
            dx = state[edge['nodes'][0]]['length'] / 2 + state[edge['nodes'][1]]['length'] / 2
            diffusion_constants = array_from(self.diffusion_constants[edge_id])
            alpha = diffusion_constants * (cross_sectional_area / dx) * timestep
            M[:, node_index_1, node_index_1] += alpha / vol_1
            M[:, node_index_2, node_index_2] += alpha / vol_2
            M[:, node_index_1, node_index_2] -= alpha / vol_1
            M[:, node_index_2, node_index_1] -= alpha / vol_2

        # Calculates final concentration after one timestep
        c_initial = np.asarray(
            [np.multiply(array_from(state[node]['molecules']),
                         array_from(self.mw)) / state[node]['volume']
             for node in state])
        c_final = np.asarray([np.matmul(np.linalg.inv(a), c_initial[:, i])
                              for i, a in enumerate(M)]).T

        # Calculates final counts
        volumes = np.asarray(
            [state[node]['volume'] for node in state])
        count_initial = np.asarray([array_from(state[node]['molecules'])
                                    for node in state])
        count_final_unrounded = np.asarray(
            [np.divide(node * volumes[i],
                       array_from(self.mw))
             for i, node in enumerate(c_final)]) + self.remainder
        count_final = np.asarray([saferound(col, 0) for col in
                                  count_final_unrounded.T]).T

        # Keeps track of remainder after rounding counts to integers
        self.remainder = count_final_unrounded - count_final
        delta = np.subtract(count_final, count_initial)

        # Ensures conservation of molecules
        assert (np.array_equal(np.ndarray.sum(count_initial, axis=0),
                np.ndarray.sum(count_final, axis=0))), 'Molecule count is not conserved'

        update = {
            node_id: {
                'molecules': array_to(self.molecule_ids,
                                      delta[np.where(self.nodes ==
                                                     node_id)[0][0]].astype(int)),
            } for node_id in self.nodes
        }
        return update


# TODO: change this to multiple tests and add asserts
def test_diffusion_network_process(out_dir=None):
    # initialize the process by passing initial_parameters
    n = int(1E6)
    molecule_ids = [str(np.round(i, 1)) for i in np.arange(0.1, 19.6, 0.1)]
    initial_parameters = {
        'nodes': ['cytosol_front', 'nucleoid', 'cytosol_rear'],
        'edges': {
            '1': {
                'nodes': ['cytosol_front', 'nucleoid'],
                'cross_sectional_area': np.pi * 0.3 ** 2,
                'mesh': True,
            },
            '2': {
                'nodes': ['nucleoid', 'cytosol_rear'],
                'cross_sectional_area': np.pi * 0.3 ** 2,
                'mesh': True,
            },
            '3': {
                'nodes': ['cytosol_front', 'cytosol_rear'],
                'cross_sectional_area': np.pi * 0.3 ** 2,
            },
            },

        'mw': {str(np.round(i, 1)): np.round(i, 1) for i in np.arange(0.1, 19.6, 0.1)},
        'mesh_size': 50,
        'radii': {str(np.round(i, 1)): np.round(i, 1) for i in np.arange(0.1, 19.6, 0.1)},
    }

    diffusion_network_process = DiffusionNetwork(initial_parameters)

    # run the simulation
    sim_settings = {
        'total_time': 10,
        'initial_state': {
            'cytosol_front': {
                'length': 0.5,
                'volume': 0.25,
                'molecules': {
                    mol_id: n
                    for mol_id in molecule_ids}
            },
            'nucleoid': {
                'length': 1.0,
                'volume': 0.5,
                'molecules': {
                    mol_id: 0
                    for mol_id in molecule_ids}
            },
            'cytosol_rear': {
                'length': 0.5,
                'volume': 0.25,
                'molecules': {
                    mol_id: 0
                    for mol_id in molecule_ids}
            },
        },
    }

    output = simulate_process_in_experiment(diffusion_network_process, sim_settings)
    rp = diffusion_network_process.rp
    diffusion_constants = diffusion_network_process.diffusion_constants

    if out_dir:
        # plot the simulation output
        plot_output(output, sim_settings['initial_state'], out_dir)
        # plot_diff_range(diffusion_constants, rp, out_dir)


# Plots the diffusion constants by molecule sizes for edges with and without mesh
def plot_diff_range(diffusion_constants, rp, out_dir='out'):
    plt.figure()
    plt.plot(np.multiply(rp, 2), array_from(diffusion_constants['1']), color='#d8b365')
    plt.plot(np.multiply(rp, 2), array_from(diffusion_constants['3']), color='#5ab4ac')
    plt.yscale('log')
    plt.xlabel(r'Molecule size ($nm$)')
    plt.ylabel(r'Diffusion constant ($\mu m^2/s$)')
    plt.title('Diffusion constants of molecules')
    plt.legend(['with 50 nm mesh', 'without mesh'])
    out_file = out_dir + '/diffusion_constants.png'
    plt.tight_layout()
    plt.savefig(out_file, dpi=300)


# Plots the normalized concentrations for the largest and smallest molecules
def plot_output(output, nodes, out_dir='out'):
    plt.figure()
    colors = ['#d8b365', '#5ab4ac', '#018571']
    large_total = array_from(
        output['cytosol_front']['molecules'])[-1][0] + array_from(
        output['nucleoid']['molecules'])[-1][0] + array_from(
        output['cytosol_rear']['molecules'])[-1][0]
    small_total = array_from(
        output['cytosol_front']['molecules'])[0][0] + array_from(
        output['nucleoid']['molecules'])[0][0] + array_from(
        output['cytosol_rear']['molecules'])[0][0]
    plt.plot(output['time'], np.divide(np.divide(
        array_from(output['cytosol_front']['molecules'])[-1],
        nodes['cytosol_front']['volume']), large_total), color=colors[0])
    plt.plot(output['time'], np.divide(np.divide(
        array_from(output['nucleoid']['molecules'])[-1],
        nodes['nucleoid']['volume']), large_total), color=colors[1])
    plt.plot(output['time'], np.divide(np.divide(
        array_from(output['cytosol_rear']['molecules'])[-1],
        nodes['cytosol_rear']['volume']), large_total), color=colors[2])
    plt.plot(output['time'], np.divide(np.divide(
        array_from(output['cytosol_front']['molecules'])[0],
        nodes['cytosol_front']['volume']), small_total),
        color=colors[0], linestyle='dashed')
    plt.plot(output['time'], np.divide(np.divide(
        array_from(output['nucleoid']['molecules'])[0],
        nodes['nucleoid']['volume']), small_total),
        color=colors[1], linestyle='dashed')
    plt.plot(output['time'], np.divide(np.divide(
        array_from(output['cytosol_rear']['molecules'])[0],
        nodes['cytosol_rear']['volume']), small_total),
        color=colors[2], linestyle='dashed')
    plt.xlabel('time (s)')
    plt.ylabel('Molecule counts')
    plt.title('Diffusion over compartments')
    plt.legend(['Cytosol front: large molecule', 'Nucleoid: large molecule',
                'Cytosol rear: large molecule', 'Cytosol front: small molecule',
                'Nucleoid: small molecule', 'Cytosol rear: small molecule'])
    out_file = out_dir + '/diffusion_large_small.png'
    plt.savefig(out_file)


# This function is modified from spatial_tool.py from WCM
def calculate_rp_from_mw(molecule_ids, mw):
    '''
        This function compute the hydrodynamic radius of a macromolecules from
        its molecular weight. It is important to note that the hydrodynamic
        diameter is mainly used for computation of diffusion constant, and can
        be different from the observed diameter under microscopes or the radius
        of gyration, especially for loose polymers such as RNAs. This function
        is not E coli specific.

        References: Bioinformatics (2012). doi:10.1093/bioinformatics/bts537

        Args:
            molecule_ids: List of molecule ids.
            mw: molecular weight of the macromolecules, units: fg.

        Returns: the hydrodynamic radius (in unit of nm) of the macromolecules
        using the following formula:
            - rp = 0.0515*MW^(0.392) nm (Hong & Lei 2008) (protein)

        These parameters are also possible for other macromolecule types,
        however all molecules are currently assumed to be proteins.
            - rp = 0.0566*MW^(0.38) nm (Werner 2011) (RNA)
            - rp = 0.024*MW^(0.57) nm (Robertson et al 2006) (linear DNA)
            - rp = 0.0125*MW^(0.59) nm (Robertson et al 2006) (circular DNA)
            - rp = 0.0145*MW^(0.57) nm (Robertson et al 2006) (supercoiled DNA)
        '''

    dic_rp = {'protein': (0.0515, 0.392),
              'RNA': (0.0566, 0.38),
              'linear_DNA': (0.024, 0.57),
              'circular_DNA': (0.0125, 0.59),
              'supercoiled_DNA': (0.0145, 0.57),
              }

    r_p0, rp_power = dic_rp['protein']
    fg_to_kDa = 602217364.34

    mw_subset = {
        key: value for key, value in mw.items() if key in molecule_ids
    }

    r_p = np.multiply(
            r_p0, np.power(np.multiply(array_from(mw_subset), fg_to_kDa), rp_power))

    return r_p


# This function is modified from spatial_tool.py from WCM
def compute_diffusion_constants_from_rp(molecule_ids, rp, mesh_size, edges,
                                       temp):
    '''
        Warning: The default values of the 'parameters' are E coli specific.

        This function computes the hypothesized diffusion constant of
        macromolecules within the nucleoid and the cytoplasm region.
        In literature, there is no known differentiation between the diffusion
        constant of a molecule in the nucleoid and in the cytoplasm up to the
        best of our knowledge in 2020. However, there is a good reason why we
        can assume that previously reported diffusion constant are in fact the
        diffusion constant of a protein in the nucleoid region:
        (1) The image traces of a protein within a bacteria usually cross the
            nucleoid regions.
        (2) The nucleoid region, compared to the cytoplasm, should be the main
        limiting factor restricting the magnitude of diffusion constant.
        (3) The same theory of diffusion constant has been implemented to
        mammalian cells, and the term 'rh', the average hydrodynamic radius of
        the biggest crowders, are different in mammalian cytoplasm, and it seems
        to reflect the hydrodynamic radius of the actin filament (note: the
        hydrodynamic radius of actin filament should be computed based on the
        average length of actin fiber, and is not equal to the radius of the
        actin filament itself.) (ref: Nano Lett. 2011, 11, 2157-2163).
        As for E coli, the 'rh' term = 40nm, which may correspond to the 80nm
        DNA fiber. On the other hand, for the diffusion constant of E coli in
        the true cytoplasm, we will expect the value of 'rh' term to be
        approximately 10 nm, which correspond to the radius of active ribosomes.

        Using these terms for scaling a baseline diffusion constant (calculated
        from Enstein-Stokes equation), a cytosol-specific diffusion calculation
        can be obtained).

        Ref: Kalwarczyk, T., Tabaka, M. & Holyst, R.
        Bioinformatics (2012). doi:10.1093/bioinformatics/bts537

        This function computes the hypothesized diffusion constant of
        macromolecules within the nucleoid region by scaling the hypothesized
        cytoplasm diffusion constant. There is evidence that as molecules
        grow in size, they become more excluded from E. coli's nucleoid because
        the DNA polymers form a meshgrid with an estimated mesh size of 50
        nm (ref: Xiang et al., bioRxiv (2020)). The equation implemented here
        assumes that molecules move through the network by encountering openings
        between the DNA meshgrid greater than the hydrodynamic radius.

        Ref: Brian Amsden
        Macromolecules (1999). doi:10.1021/ma980922a

        D_0 = K_B*T/(6*pi*eta_0*rp)
        ln(D_0/D_cyto) = ln(eta/eta_0) = (xi^2/Rh^2 + xi^2/rp^2)^(-a/2)
        D_0 = the diffusion constant of a macromolecule in pure solvent
        eta_0 = the viscosity of pure solvent, in this case, water
        eta = the size-dependent viscosity experienced by the macromolecule.
        xi = average distance between the surface of proteins
        rh = average hydrodynamic radius of the biggest crowders
        a = some constant of the order of 1
        rp = hydrodynamic radius of probed molecule

        In this formula, since we allow the changes in temperature, we also
        consider the viscosity changes of water under different temperature:
        Ref: Dortmund Data Bank
        eta_0 = A*10^(B/(T-C))
        A = 2.414*10^(-5) Pa*sec
        B = 247.8 K
        C = 140 K

        Args:
            molecule_ids: List of molecule ids.
            rp: List of radii corresponding to each molecule id. unit: nm
            mesh_size: Size of meshgrid openings. unit: nm
            edges: Dictionary of edges.
            temp: The temperature of interest. unit: K.

        Returns:
            dc: the diffusion constant of the macromolecule, units: um**2/sec
        '''
    if temp is None:
        temp = 310.15
    parameters = (0.51, 0.53, 10)

    # unpack constants required for the calculation
    xi, a, rh_cyto = parameters  # unit: nm, 1, nm

    ro = mesh_size/2
    rh = rh_cyto
    diffusion_constants = {}
    K_B = scipy.constants.Boltzmann  # Boltzmann constant, unit: J/K

    # calculate viscosity of water based on temperature
    a_visc = 2.414*10**(-5)  # unit: Pa*sec
    b_visc = 247.8  # unit: K
    c_visc = 140  # unit: K
    eta_0 = a_visc*10**(b_visc/(temp - c_visc))  # unit: Pa*sec

    # conversions
    m2_to_um2 = 1E12
    nm_to_m = 1E-9

    # compute dc (diffusion constant)
    dc_0 = np.multiply(
        np.divide(K_B*temp, np.multiply(6*np.pi*eta_0*nm_to_m, rp)), m2_to_um2)
    dc_cyto = np.multiply(dc_0, np.exp(
        -np.power(np.add(xi**2/rh**2, np.divide(xi**2, np.square(rp))), (-a/2))))

    # compute impact from mesh when nucleoid is a node
    dc_nuc = np.multiply(
        dc_cyto, np.exp(np.multiply((-np.pi / 4), np.square(np.divide(rp, ro)))))

    for edge_id, edge in edges.items():
        scaling_factor = 1
        mesh = False
        if 'mesh' in edge:
            mesh = edge['mesh']
        if 'diffusion_scaling_constant' in edge:
            scaling_factor = edge['diffusion_scaling_constant']
        if mesh:
            diffusion_constants[edge_id] = array_to(
                molecule_ids, dc_nuc * scaling_factor)
        else:
            diffusion_constants[edge_id] = array_to(
                molecule_ids, dc_cyto * scaling_factor)
        if 'diffusion_constants' in edge:
            for mol_id, dc in edge['diffusion_constants'].items():
                diffusion_constants[edge_id][mol_id] = dc

    return diffusion_constants


# Helper function
def array_from(d):
    return np.array(list(d.values()))

# Helper function
def array_to(keys, array):
    return {
        key: array[index]
        for index, key in enumerate(keys)}


# run module is run as the main program with python vivarium/process/template_process.py
if __name__ == '__main__':
    # make an output directory to save plots
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    test_diffusion_network_process(out_dir)






from __future__ import absolute_import, division, print_function

from vivarium.library.units import units

from scipy import constants

nAvogadro = constants.N_A

COUNTS_UNITS = units.mmol
VOLUME_UNITS = units.L
MASS_UNITS = units.g
TIME_UNITS = units.s
CONC_UNITS = COUNTS_UNITS / VOLUME_UNITS


def molar_to_counts(fluxes, volume):
    '''
    input:
        fluxes -- list (molar)
        volume -- list (L)

    return:
        counts -- list
    '''
    # volume = cell_mass / density
    return (nAvogadro * volume * fluxes).astype(int)

def millimolar_to_counts(fluxes, volume):
    '''
    input:
        fluxes -- list (millimolar)
        volume -- list (L)

    return:
        counts -- list
    '''
    fluxes_mol = fluxes * 1e-3 # convert to molar
    return int(nAvogadro * volume * fluxes_mol)


def counts_to_molar(counts, volume):
    '''
    input:
        counts -- list
        volume -- list (L)

    return:
        fluxes -- list (molar)
    '''
    # volume = cell_mass / density
    return counts / (nAvogadro * volume)


def counts_to_millimolar(counts, volume):
    '''
    input:
        counts -- list
        volume -- list (L)

    return:
        fluxes -- list (molar)
    '''
    # volume = cell_mass / density
    return counts / (nAvogadro * volume * 1e-3)


def molar_to_molDCWhr(fluxes, dry_mass, cell_mass, density, timestep):
    '''
    input:
        flux: mmol/L
        dry_mass: fg
        cell_mass: fg
        density:
        timestep: seconds
    return:
        fluxes in mol/gDCW/hr
    '''
    # Coefficient to convert between flux (mol/g DCW/hr) basis and concentration (M) basis
    coefficient = dry_mass / cell_mass * density * timestep
    return (fluxes / coefficient).m_as(units.mmol / units.g / units.h)


import os

from vivarium.core.process import Composite
from vivarium.core.composition import (
    simulate_compartment_in_experiment,
    COMPOSITE_OUT_DIR
)
from vivarium.plots.simulation_output import plot_simulation_output
from vivarium.plots.topology import plot_topology

# processes/composites
from vivarium_cell.plots.gene_expression import plot_gene_expression_output
from vivarium_cell.processes.division_volume import DivisionVolume
from vivarium_cell.processes.metabolism import Metabolism, get_iAF1260b_config
from vivarium_cell.processes.convenience_kinetics import ConvenienceKinetics, get_glc_lct_config
from vivarium_cell.processes.transcription import Transcription
from vivarium_cell.processes.translation import Translation
from vivarium_cell.processes.degradation import RnaDegradation
from vivarium_cell.processes.complexation import Complexation
from vivarium_cell.states.chromosome import toy_chromosome_config
from vivarium_cell.processes.local_field import LocalField


NAME = 'master'


def default_metabolism_config():
    metabolism_config = get_iAF1260b_config()
    metabolism_config.update({
        'moma': False,
        'tolerance': {
            'EX_glc__D_e': [1.05, 1.0],
            'EX_lcts_e': [1.05, 1.0]}})
    return metabolism_config



class Master(Composite):

    defaults = {
        'global_path': ('global',),
        'external_path': ('external',),
        'transport': get_glc_lct_config(),
        'metabolism': default_metabolism_config(),
        'fields_path': ('fields',),
        'dimensions_path': ('dimensions',),
        }

    def __init__(self, config=None):
        super(Master, self).__init__(config)

    def generate_processes(self, config):

        # Transport
        transport_config = config.get('transport')
        transport = ConvenienceKinetics(transport_config)
        target_fluxes = transport.kinetic_rate_laws.reaction_ids

        # Metabolism
        # add target fluxes from transport
        metabolism_config = config.get('metabolism')
        metabolism_config.update({'constrained_reaction_ids': target_fluxes})
        metabolism = Metabolism(metabolism_config)

        local_field = LocalField()

        # Expression
        transcription_config = config.get('transcription', {})
        translation_config = config.get('translation', {})
        degradation_config = config.get('degradation', {})
        transcription = Transcription(transcription_config)
        translation = Translation(translation_config)
        degradation = RnaDegradation(degradation_config)
        complexation = Complexation(config.get('complexation', {}))

        # Division
        # get initial volume from metabolism
        division_config = config.get('division', {})
        division_config.update({'initial_state': metabolism.initial_state()})
        division = DivisionVolume(division_config)

        return {
            'metabolism': metabolism,
            'transport': transport,
            'local_field': local_field,
            'transcription': transcription,
            'translation': translation,
            'degradation': degradation,
            'complexation': complexation,
            'division': division,
        }

    def generate_topology(self, config):
        global_path = config['global_path']
        external_path = config['external_path']
        fields_path = config['fields_path']
        dimensions_path = config['dimensions_path']
        return {
            'transport': {
                'internal': ('metabolites',),
                'external': external_path,
                'exchanges': ('null',),  # metabolism's exchange is used
                'fluxes': ('flux_bounds',),
                'global': global_path,
            },
            'metabolism': {
                'internal': ('metabolites',),
                'external': external_path,
                'reactions': ('reactions',),
                'exchanges': ('exchanges',),
                'flux_bounds': ('flux_bounds',),
                'global': global_path,
            },
            'local_field': {
                'exchanges': ('exchanges',),
                'fields': fields_path,
                'location': global_path + ('location',),
                'dimensions': dimensions_path,
            },
            'transcription': {
                'chromosome': ('chromosome',),
                'molecules': ('metabolites',),
                'proteins': ('proteins',),
                'transcripts': ('transcripts',),
                'factors': ('concentrations',),
                'global': global_path
            },
            'translation': {
                'ribosomes': ('ribosomes',),
                'molecules': ('metabolites',),
                'transcripts': ('transcripts',),
                'proteins': ('proteins',),
                'concentrations': ('concentrations',),
                'global': global_path
            },
            'degradation': {
                'transcripts': ('transcripts',),
                'proteins': ('proteins',),
                'molecules': ('metabolites',),
                'global': global_path
            },
            'complexation': {
                'monomers': ('proteins',),
                'complexes': ('proteins',),
                'global': global_path
            },
            'division': {
                'global': global_path
            }
        }


def run_master(out_dir):
    timeseries = test_master()
    volume_ts = timeseries['global']['volume']
    print('growth: {}'.format(volume_ts[-1]/volume_ts[0]))
    expression_plot_settings = {
        'name': 'gene_expression',
        'ports': {
            'transcripts': 'transcripts',
            'molecules': 'metabolites',
            'proteins': 'proteins'}}
    plot_gene_expression_output(timeseries, expression_plot_settings, out_dir)

    plot_settings = {
        'max_rows': 20,
        'remove_zeros': True,
        'skip_ports': ['prior_state', 'null', 'flux_bounds', 'chromosome', 'reactions']}
    plot_simulation_output(timeseries, plot_settings, out_dir)

def load_master():
    # load the compartment
    config = {
        'external_path': ('external',),
        'global_path': ('global',),
        'agents_path': ('..', '..', 'cells',),
        'transcription': {
            'sequence': toy_chromosome_config['sequence'],
            'templates': toy_chromosome_config['promoters'],
            'genes': toy_chromosome_config['genes'],
            'promoter_affinities': toy_chromosome_config['promoter_affinities'],
            'transcription_factors': ['tfA', 'tfB'],
            'elongation_rate': 10.0}}
    return Master(config)


def test_master():
    master = load_master()
    # simulate
    settings = {
        'timestep': 1,
        'total_time': 10}
    return simulate_compartment_in_experiment(master, settings)


def master_topology(out_dir='out'):
    master = load_master()
    plot_topology(master, out_dir=out_dir, filename='master_topology')

if __name__ == '__main__':
    out_dir = os.path.join(COMPOSITE_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    master_topology(out_dir)
    run_master(out_dir)



'''
===============================================
Ordinary Differential Equation Expression Model
===============================================
'''

from __future__ import absolute_import, division, print_function

import os
import argparse
import random
import math
import logging as log

from vivarium.core.process import Process
from vivarium.library.dict_utils import deep_merge, tuplify_port_dicts
from vivarium.core.composition import (
    simulate_process_in_experiment,
    PROCESS_OUT_DIR,
)
from vivarium.plots.simulation_output import plot_simulation_output
from vivarium.library.units import units
from vivarium_cell.library.regulation_logic import build_rule
from vivarium_cell.processes.derive_globals import AVOGADRO


NAME = 'ode_expression'

class ODE_expression(Process):
    '''Models gene expression using ordinary differential equations

     This :term:`process class` models the rates of transcription,
     translation, and degradation using ordinary differential
     equations (ODEs). These ODEs are as follows:

     * Transcription: :math:`\\frac{dM}{dt} = k_M - d_M M`

         * :math:`M`: Concentration of mRNA
         * :math:`k_M`: Trancription rate
         * :math:`d_M`: Degradation rate

     * Translation: :math:`\\frac{dP}{dt} = k_P m_P - d_P P`

         * :math:`P`: Concentration of protein
         * :math:`m_P`: Concentration of transcript
         * :math:`k_P`: Translation rate
         * :math:`d_P`: Degradation rate

     The transcription rate abstracts over factors that include gene
     copy number, RNA polymerase abundance, promoter strengths, and
     nucleotide availability. The translation rate similarly
     abstracts over factors that include ribosome availability, the
     binding strength of the mRNA to the ribosome, the availability
     of tRNAs, and the availability of free amino acids.

     This process class also models genetic regulation with boolean
     logic statements. This restricts us to modeling binary
     regulation: reactions can be completely suppressed, but they
     cannot be only slowed. For example, we can model a statement
     like:

         If :math:`[glucose] > 0.1`, completely inhibit LacY
         expression.

     But we cannot model a statement like this:

         If :math:`[glucose] > 0.1`, reduce LacY expression by 50%.

     .. note::
         This model does **not** include:

             * Kinetic regulation
             * Cooperativity
             * Autoinhibition
             * Autoactivation

     :term:`Ports`:

     * **internal**: Expects a :term:`store` with all of the
       transcripts and proteins being modeled. This store should also
       include any internal regulators.
     * **external**: Expects a store with all external regulators.
       These variables are not updated; they are only read.

     Arguments:
         initial_parameters: A dictionary of configuration options.
             The following configuration options may be provided:

             * **transcription_rates** (:py:class:`dict`): Maps
               transcript names (the keys of the dict) to
               transcription rates (values of the dict).
             * **translation_rates** (:py:class:`dict`): Maps protein
               names (the keys of the dict) to translation rates (the
               values of the dict).
             * **degradation_rates** (:py:class:`dict`): Maps from
               names of substrates (transcripts or proteins) to
               degrade to their degradation rates.
             * **protein_map** (:py:class:`dict`): Maps from protein
               name to transcript name for each transcript-protein
               pair.
             * **initial_state** (:py:class:`dict`): Maps from
               :term:`port` names to the initial state of that port.
               Each initial state is specified as a dictionary with
               :term:`variable` names as keys and variable values as
               values.
             * **regulators** (:py:class:`list`): List of the
               molecules that regulate any of the modeled
               transcription or translation reactions. Each molecule
               is a tuple of the port and the molecule's variable
               name.
             * **regulation** (:py:class:`dict`): Maps from reaction
               product (transcript or protein) name to a boolean
               regulation statement. For example, to only transcribe
               ``lacy_RNA`` if external ``glc__D_e`` concentration is
               at most 0.1, we would pass:

               .. code-block:: python

                 {'lacy_RNA': 'if not (external, glc__D_e) > 0.1'}

     Example instantiation of a :term:`process` modeling LacY
     expression:

     >>> initial_state = {
     ...     'internal': {
     ...         'lacy_RNA': 0.0,
     ...         'LacY': 0.0,
     ...     },
     ...     'external': {
     ...         'glc__D_e': 2.0,
     ...     }
     ... }
     >>> config = {
     ...     'transcription_rates': {
     ...         'lacy_RNA': 3.0,
     ...     },
     ...     'translation_rates': {
     ...         'LacY': 4.0,
     ...     },
     ...     'degradation_rates': {
     ...         'lacy_RNA': 1.0,
     ...         'LacY': 0.0,
     ...     },
     ...     'protein_map': {
     ...         'LacY': 'lacy_RNA',
     ...     },
     ...     'initial_state': initial_state,
     ...     'regulators': [('external', 'glc__D_e')],
     ...     'regulation': {
     ...         'lacy_RNA': 'if (external, glc__D_e) > 1.0',
     ...     },
     ... }
     >>> expression_process = ODE_expression(config)
     >>> state = expression_process.default_state()
     >>> state == initial_state
     True
     >>> # When external glc__D_e present, no transcription
     >>> expression_process.next_update(1, state)
     {'internal': {'lacy_RNA': 0.0, 'LacY': 0.0}}
     >>> # But translation and degradation still occur
     >>> state['internal']['lacy_RNA'] = 1.0
     >>> expression_process.next_update(1, state)
     {'internal': {'lacy_RNA': -1.0, 'LacY': 4.0}}
     >>> # When external glc__D_e low, LacY transcribed
     >>> state['external']['glc__D_e'] = 0.5
     >>> expression_process.next_update(1, state)
     {'internal': {'lacy_RNA': 2.0, 'LacY': 4.0}}

     '''

    name = NAME
    defaults = {
        'time_step': 1.0,
        'transcription_rates': {},
        'translation_rates': {},
        'degradation_rates': {},
        'protein_map': {},
        'transcription_leak': {
            'rate': 0.0,  # probability of leak in 1 second
            'magnitude': 0.0,  # the amount of leak
        },
        'regulation': {},
        'regulators': [],
        'initial_state': {},
        'counts_deriver_key': 'expression_counts',
    }

    def __init__(self, parameters=None):
        super(ODE_expression, self).__init__(parameters)

        # ode gene expression
        self.transcription = self.parameters.get('transcription_rates')
        self.translation = self.parameters.get('translation_rates')
        self.degradation = self.parameters.get('degradation_rates')
        self.protein_map = self.parameters.get('protein_map')
        transcription_leak = self.parameters.get('transcription_leak')
        self.transcription_leak_rate = transcription_leak['rate']
        self.transcription_leak_magnitude = transcription_leak['magnitude']

        # boolean regulation
        regulation_logic = self.parameters.get('regulation')
        self.regulation = {
            gene_id: build_rule(logic) for gene_id, logic in regulation_logic.items()}
        regulators = self.parameters.get('regulators')
        self.internal_regulators = [state_id for port_id, state_id in regulators if port_id == 'internal']
        self.external_regulators = [state_id for port_id, state_id in regulators if port_id == 'external']

        # get initial state
        states = list(self.transcription.keys()) + list(self.translation.keys())
        null_states = {'internal': {
            state_id: 0 for state_id in states}}
        initialized_states = self.parameters.get('initial_state')
        self.initial_state = deep_merge(null_states, initialized_states)
        self.internal = list(self.initial_state.get('internal', {}).keys())
        self.external = list(self.initial_state.get('external', {}).keys())
        self.counts_deriver_key = self.parameters.get('counts_deriver_key')


    def ports_schema(self):
        ports = [
            'internal',
            'external',
            'counts',
            'global',
        ]
        schema = {port: {} for port in ports}

        # internal
        for state in self.internal + self.internal_regulators:
            schema['internal'][state] = {
                '_default': self.initial_state['internal'].get(state, 0.0),
                '_emit': True,
            }

        # external
        for state in self.external + self.external_regulators:
            schema['external'][state] = {
                '_default': self.initial_state['external'].get(state, 0.0),
                '_emit': True,
            }

        # counts
        for state in self.internal + self.internal_regulators:
            schema['counts'][state] = {
                '_divider': 'split',
                '_emit': True
            }

        # global
        schema['global'] = {}

        return schema

    def derivers(self):
        return {
            self.counts_deriver_key: {
                'deriver': 'counts_deriver',
                'port_mapping': {
                    'global': 'global',
                    'concentrations': 'internal',
                    'counts': 'counts'},
                'config': {
                    'concentration_keys': self.internal + self.internal_regulators}}}

    def next_update(self, timestep, states):
        internal_state = states['internal']

        # get current condition of regulated genes (True/False)
        flattened_states = tuplify_port_dicts(states)
        regulation_condition = {}
        for gene_id, reg_logic in self.regulation.items():
            regulation_condition[gene_id] = reg_logic(flattened_states)

        internal_update = {}
        # transcription: dM/dt = k_M - d_M * M
        # M: conc of mRNA, k_M: transcription rate, d_M: degradation rate
        for transcript, rate in self.transcription.items():
            # do not transcribe regulated genes (rate = 0), except for transcriptional leaks
            if regulation_condition.get(transcript, False):
                # leak probability as function of the time step
                leak_rate = -math.log(1 - self.transcription_leak_rate)
                leak_probability = 1 - math.exp(-leak_rate * timestep)
                if random.uniform(0, 1) < leak_probability:
                    rate = self.transcription_leak_magnitude
                else:
                    rate = 0.0

            transcript_state = internal_state[transcript]
            internal_update[transcript] = \
                (rate - self.degradation.get(transcript, 0) * transcript_state) * timestep

        # translation: dP/dt = k_P * m_P - d_P * P
        # P: conc of protein, m_P: conc of P's transcript, k_P: translation rate, d_P: degradation rate
        for protein, rate in self.translation.items():
            transcript = self.protein_map[protein]
            transcript_state = internal_state[transcript]
            protein_state = internal_state[protein]

            internal_update[protein] = \
                (rate * transcript_state - self.degradation.get(protein, 0) * protein_state) * timestep

        return {'internal': internal_update}


# test functions
# toy config
def get_lacy_config():
    '''LacY ODE expresion configuration

    This configuration can be passed to :py:class:`ODE_expression` like
    this:

    >>> config = get_lacy_config()
    >>> expression_process = ODE_expression(config)
    '''
    transcription_rates = {
        'lacy_RNA': 1e-7}

    translation_rates = {
        'LacY': 5e-3}

    protein_map = {
        'LacY': 'lacy_RNA'}

    degradation_rates = {
        'lacy_RNA': 3e-3,  # a single RNA lasts about 5 minutes
        'LacY': 3e-5}

    # define regulation
    regulators = [
        ('external', 'glc__D_e'),
        ('internal', 'lcts_p')]
    regulation_condition = {
        'lacy_RNA': 'if [(external, glc__D_e) > 0.005 '  # limiting concentration of glc at 0.005 mM (Boulineau 2013)
                    'or (internal, lcts_p) < 0.005]'  # internal lcts is hypothesized to disinhibit lacY transcription
    }
    transcription_leak = {
        'rate': 1e-4,
        'magnitude': 1e-6,
        }

    # initial state
    initial_state = {
        'internal': {
            'lacy_RNA': 0.0,
            'LacY': 0.0},
        'external': {
            'glc__D_e': 8.0,
            'lcts_e': 8.0}}

    return {
        'transcription_rates': transcription_rates,
        'translation_rates': translation_rates,
        'degradation_rates': degradation_rates,
        'protein_map': protein_map,
        'regulators': regulators,
        'regulation': regulation_condition,
        'transcription_leak': transcription_leak,
        'initial_state': initial_state}

def get_flagella_expression():
    '''Flagella ODE expresion configuration

    This configuration can be passed to :py:class:`ODE_expression` like
    this:

    >>> config = get_flagella_expression()
    >>> expression_process = ODE_expression(config)
    '''
    transcription = {
        'flag_RNA': 1e-6}

    translation = {
        'flagella': 8e-5}

    degradation = {
        'flag_RNA': 2e-2}  # 1e-23}

    protein_map = {
        'flagella': 'flag_RNA'}

    # get initial concentrations from counts
    volume = 1.2 * units.fL
    mmol_to_counts = (AVOGADRO * volume).to('L/mmol')
    counts = {
        'flagella': 0,
        'flag_RNA': 0}
    concentrations = {}
    for state_id, count in counts.items():
        concentrations[state_id] = (count / mmol_to_counts).magnitude

    initial_state = {
        'counts': counts,
        'internal': concentrations}

    return {
        'transcription_rates': transcription,
        'translation_rates': translation,
        'degradation_rates': degradation,
        'protein_map': protein_map,
        'initial_state': initial_state}


def test_expression(config=get_lacy_config(), timeline=[(100, {})]):
    expression = ODE_expression(config)
    settings = {'timeline': {'timeline': timeline}}
    return simulate_process_in_experiment(expression, settings)


if __name__ == '__main__':
    out_dir = os.path.join(PROCESS_OUT_DIR, NAME)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    parser = argparse.ArgumentParser(description='ODE expression')
    parser.add_argument('--lacY', '-l', action='store_true', default=False)
    parser.add_argument('--flagella', '-f', action='store_true', default=False)
    args = parser.parse_args()

    if args.flagella:
        timeline = [(2520, {})]
        timeseries = test_expression(get_flagella_expression(), timeline)
        plot_simulation_output(timeseries, {}, out_dir, 'flagella_expression')
    else:
        total_time = 5000
        shift_time1 = int(total_time / 5)
        shift_time2 = int(3 * total_time / 5)
        timeline = [
            (0, {('external', 'glc__D_e'): 10}),
            (shift_time1, {('external', 'glc__D_e'): 0}),
            (shift_time2, {('external', 'glc__D_e'): 10}),
            (total_time, {})]
        timeseries = test_expression(get_lacy_config(), timeline)
        plot_simulation_output(timeseries, {}, out_dir, 'lacY_expression')



'''
==================================
Utilities for Lattice Environments
==================================
'''

from __future__ import absolute_import, division, print_function

import numpy as np
from scipy import constants

from vivarium.library.units import units, Quantity


AVOGADRO = constants.N_A / units.mol


def get_bin_site(location, n_bins, bounds):
    '''Get a bin's indices in the lattice

    Parameters:
        location (list): A list of 2 floats that specify the x and y
            coordinates of a point inside the desired bin.
        n_bins (list): A list of 2 ints that specify the number of bins
            along the x and y axes, respectively.
        bounds (list): A list of 2 floats that define the dimensions of
            the lattice environment along the x and y axes,
            respectively.

    Returns:
        tuple: A 2-tuple of the x and y indices of the bin in the
        lattice.
    '''
    bin_site_no_rounding = np.array([
        location[0] * n_bins[0] / bounds[0],
        location[1] * n_bins[1] / bounds[1]
    ])
    bin_site = tuple(
        np.floor(bin_site_no_rounding).astype(int) % n_bins)
    return bin_site


def get_bin_volume(n_bins, bounds, depth):
    '''Get a bin's volume

    Parameters:
        n_bins (list): A list of 2 ints that specify the number of bins
            along the x and y axes, respectively.
        bounds (list): A list of 2 floats that specify the lengths of
            the environment's sides in the x and y directions,
            respectively. In units of microns.
        depth (float): The depth of the environment, in microns.

    Returns:
        float: The volume of each bin in the lattice, in Liters.
    '''
    total_volume = (depth * bounds[0] * bounds[1]) * 1e-15  # (L)
    return total_volume / (n_bins[0] * n_bins[1])


def count_to_concentration(count, bin_volume):
    '''Convert a molecule count into a concentration.

    Parameters should all have units. Returned value will have units.

    Parameters:
        count (int): The number of molecules in the bin.
        bin_volume (float): The volume of the bin.

    Returns:
        float: The concentration of molecule in the bin.
    '''
    return count / (bin_volume * AVOGADRO)
